# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from _annotation_tools.xml_structured_parsing.myXML import myXML, DELIMITER, ET

#def parser_xml_CyberPSG(path):
#    class EmptyClass:
#        def __init__(self, version=1.1, file_id='SPARSE_MIX', file_path=''):
#            self._version = version
#            self._file_id = file_id
#            self._file_path = file_path
#
#    PseudoParent = EmptyClass(version=1.1, file_id=path.split(DELIMITER)[-1].split('.')[0], file_path=path)
#    return myXML_AnnotationData.init_from_existing(PseudoParent, ET.parse(path).getroot())

##### VERSION #####
class myXML_Note(myXML):
    tag = 'note'

class myXML_Id(myXML):
    tag = 'id'

##### META #####
class myXML_Created(myXML):
    tag = 'created'

class myXML_Modified(myXML):
    tag = 'modified'

class myXML_Name(myXML):
    tag = 'name'

class myXML_AnnotationGroup(myXML):
    tag = 'AnnotationGroup'
    ref_children = [myXML_Id, myXML_Created, myXML_Modified, myXML_Name, myXML_Note]

class myXML_AnnotationGroups(myXML):
    tag = 'AnnotationGroups'
    ref_children = [myXML_AnnotationGroup]

class myXML_Description(myXML):
    tag = 'description'

class myXML_stdDurationInSec(myXML):
    tag = 'stdDurationInSec'

class myXML_startsWithEpoch(myXML):
    tag = 'startsWithEpoch'

class myXML_Hotkey(myXML):
    tag = 'hotkey'

class myXML_Color(myXML):
    tag = 'color'

class myXML_GroupAsociations(myXML):
    tag = 'groupAssociations'
    ref_children = [myXML_Id]

class myXML_AnnotationType(myXML):
    tag = 'AnnotationType'
    ref_children = [myXML_Id, myXML_Created, myXML_Modified, myXML_Name, myXML_Description, myXML_stdDurationInSec, myXML_startsWithEpoch, myXML_Hotkey, myXML_Color, myXML_GroupAsociations, myXML_Note]

class myXML_AnnotationTypes(myXML):
    tag = 'AnnotationTypes'
    ref_children = [myXML_AnnotationType]

class myXML_StartTimeUtc(myXML):
    tag = 'startTimeUtc'

class myXML_endTimeUtc(myXML):
    tag = 'endTimeUtc'

class myXML_AnnotationTypeId(myXML):
    tag = 'annotationTypeId'

class myXML_ChannelName(myXML):
    tag = 'channelName'

class myXML_VerticalPositionPercentage(myXML):
    tag = 'verticalPositionPercentage'

class myXML_Annotation(myXML):
    tag = 'Annotation'
    ref_children = [myXML_Id, myXML_Created, myXML_Modified, myXML_StartTimeUtc, myXML_endTimeUtc, myXML_AnnotationTypeId, myXML_ChannelName, myXML_VerticalPositionPercentage, myXML_Note]

class myXML_Annotations(myXML):
    tag = 'Annotations'
    ref_children = [myXML_Annotation]

##### 1st Level ######
class myXML_AnnotationData(myXML):
    tag = 'AnnotationData'
    ref_children = [myXML_AnnotationGroups, myXML_AnnotationTypes, myXML_Annotations]


# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


"""
Sleep Feature Extractor
^^^^^^^^^^^^^^^^^^^^^^^

The Feature Extractor package contains SleepSpectralFeatureExtractor for spectral feature extraction from designed for sleep classification from a raw EEG signal.


The extractor can return **data rate** which gives a relative ratio of valid values in the input signal based on a number of NaN values. The extractor requires information about frequency bands at which parameters will be calculated. Please see an example bellow.



Example
^^^^^^^^^

.. code-block:: python

    import sys
    import numpy as np
    from best.feature_extraction.Feature_Extractor import SleepSpectralFeatureExtractor

    # Example synthetic signal generator
    fs = 500 # sampling frequency
    f = 10 # sin frequency
    a = 1 # amplitude
    b = 0 # bias
    t = np.arange(0, 1000, 1/fs)
    x = a * np.sin(2*np.pi*f*t) + b


    # Spectral Feature  Extraction
    fs = 500 # sampling frequency of an analysed signal
    segm_size = 30 # time length of a segment which is used for extraction of each feature
    fbands = [[1, 4],
     [4, 8],
     [8, 12],
     [12, 14],
     [14, 20],
     [20, 30]] # frequency bands at which you want to extract features

    from sleep_classification.FeatureExtractor.SpectralFeatures import mean_bands, relative_bands

    Extractor = SleepSpectralFeatureExtractor(
        fs=fs,
        segm_size=segm_size,
        fbands=fbands,
        datarate=True
    )


    Extractor_MeanBand._extraction_functions = \
        [
            mean_bands,
            mean_frequency,
            relative_bands,
        ]

    feature_values, feature_names = Extractor(x)



Sources
^^^^^^^^^^^

This Feature Extractor implementation is based on the following papers (when use whole, parts, or are inspired by, we appreciate you acknowledge and refer these journal papers)


| Kremen, V., Duque, J. J., Brinkmann, B. H., Berry, B. M., Kucewicz, M. T., Khadjevand, F., G.A. Worrell, G. A. (2017). Behavioral state classification in epileptic brain using intracranial electrophysiology. Journal of Neural Engineering, 14(2), 026001. https://doi.org/10.1088/1741-2552/aa5688


| Kremen, V., Brinkmann, B. H., Van Gompel, J. J., Stead, S. (Matt) M., St Louis, E. K., & Worrell, G. A. (2018). Automated Unsupervised Behavioral State Classification using Intracranial Electrophysiology. Journal of Neural Engineering. https://doi.org/10.1088/1741-2552/aae5ab


| Gerla, V., Kremen, V., Macas, M., Dudysova, D., Mladek, A., Sos, P., & Lhotska, L. (2019). Iterative expert-in-the-loop classification of sleep PSG recordings using a hierarchical clustering. Journal of Neuroscience Methods, 317(February), 61?70. https://doi.org/10.1016/j.jneumeth.2019.01.013


and on repository `Semi Automated Sleep Classifier <https://github.com/vkremen/Semi_Automated_Sleep_Classifier_iEEG>`_, see details in the original repository.


"""

import multiprocessing

import numpy as np
import scipy.signal as signal

from functools import partial

from best.types import ObjDict
from best.feature_extraction.SpectralFeatures import *
from best.signal import buffer, PSD


class SleepSpectralFeatureExtractor:
    __version__ = "2.0.0"
    """
    Spectral feature extractor designed for sleep classification

    ...

    Attributes
    ----------
    extraction_functions : list
        List of functions for parameter extraction from np.ndarray where shape is (n_spectrums, spectrum_samples).
        Each function has to return tuple (features : list, name_of_features : list). By default set to extract all features within this class.

    ...

    Methods
    --------

    ...

    Notes
    ------
    v0.1.0 Updates
    - communication between functions (Pxx, fs, ...) changed to ObjDict - see AISC.types.ObjDict
    - float value frequency bands enabled
    v0.1.1 Updates
    - bands_to_erase as an input into __call__ - erases defined bands of psd
    v0.1.2 Updates
    - self._extraction_functions init moved to __init__    
    v0.2.0
        - methods buffer, PSD transferred into AISC.utils.signal
        - PSD replaced by welch's method implementation by scipy.signal.welch
        - implementation of spectral feature extraction from shorter segments and average them
        - removed filtering due switch to welch method of periodogram - constant detrending
            - can be replaced with LowFrequencyFilter later on
        - automatic windown implemented in welch method - hann window

    v1.0.0
        - implemented combining of fbands and bands to erase to get bands for subsequent filtration
        - implemented filtering of frequency bands to eliminate leakage of surrounding frequencies
        - parameter init for feature extraction was moved from __call__ and process_signal to __init__
        - feature extraction functions were moved into AISC.FeatureExtractor.SpectralFeatures

    v2.0.0
        - notch filter implemented for stim frequency attenuation
        - chebyshev (cheb2) filter implemented for filtering individual bands
        - bands_to_erase modified from list of bands to list of single values which are to be filtered using notch filter

    """

    def __init__(self,
                 fs: float = None,
                 segm_size: float = None,
                 fbands: list = None,
                 sperwelchseg: float = None,
                 soverlapwelchseg: float = 0.0,
                 nfft: int = None,
                 filters_attenuate: list = [],
                 repeat_attenuate: int = 1,
                 ignore_bands: list = [],
                 datarate: bool = False,
                 n_processes: int = 1):

        self._fs = self._verify_input_fs(fs)
        self._segm_size = self._verify_input_segm_size(segm_size)
        self._fbands = self._verify_input_fbands(fbands)

        ###### SIGNAL PREPROCESSING SETUP #########

        self.FILTERS = None
        self.FILTER_LOW = signal.butter(6, np.min(fbands), fs=fs, btype='high', analog=False)
        self.FILTER_HIGH = signal.butter(6, np.max(fbands) + 10, fs=fs, btype='low', analog=False)

        self.FILTERS_ATTENUATE = []
        if filters_attenuate == 2:
            self.FILTERS_ATTENUATE = [signal.iirnotch(2 * k, Q=(20 + k * 2) * 2, fs=fs) for k in
                                      range(1, int(np.floor(np.max(fbands) / 2)) + 1)]
        elif filters_attenuate == 7:
            self.FILTERS_ATTENUATE = [signal.iirnotch(7 * k, Q=(2 + k * 2) * 7, fs=fs) for k in
                                      range(1, int(np.floor(np.max(fbands) / 7)) + 1)]

        if isinstance(filters_attenuate, list):
            if filters_attenuate.__len__() > 0:
                self.FILTERS_ATTENUATE = filters_attenuate
        self.ATTENUATE_N = repeat_attenuate

        ##### FEATURE EXTRACTION SETUP ###########

        self._ignore_bands = ignore_bands
        self.datarate = datarate
        self.n_processes = self._verify_input_n_processes(n_processes)

        self._sperwelchseg = sperwelchseg
        self._soverlapwelchseg = soverlapwelchseg
        self._nfft = nfft

        # self.extraction_functions = [
        #    normalized_entropy, mean_frequency, median_frequency, mean_bands, relative_bands, normalized_entropy_bands
        # ]  # property setter

        self.extraction_functions = [
            mean_bands, relative_bands
        ]  # property setter

    def design_filters(self):
        if self._filter_bands:
            self.FILTERS = []
            self.FILTER_RESPONSE = None
            trans = None
            for band in self._filter_bands:
                cutoff_low = 2 * band[1] / self._fs
                cutoff_high = 2 * band[0] / self._fs

                b_low = signal.firwin(self._nfiltorder, cutoff_low, pass_zero='lowpass')
                if band[0] > 0:
                    b_high = signal.firwin(self._nfiltorder, cutoff_high, pass_zero='highpass')
                    b = signal.convolve(b_low, b_high, mode='same')
                else:
                    b = cutoff_low

                w, h = signal.freqz(b, [1])
                b /= h.__abs__().max()

                self.FILTERS += [b]

                w, h = signal.freqz(b, [1])
                w = w / w.max()
                w = w * 0.5 * self._fs

                if isinstance(trans, type(None)):
                    trans = h
                else:
                    trans = trans + h
            self.FILTER_RESPONSE = {'x': w, 'y': trans}

    def __call__(
            self,
            x: (np.ndarray, list) = None
    ):

        """

        Parameters
        ----------
        x : np.ndarray
        fs
        segm_size
        sub_segment_size
        sub_segment_overlap
        fbands
        bands_to_erase
        datarate
        n_processes

        Returns
        -------

        """
        # Standard parameters
        if isinstance(x, np.ndarray):
            return self.process_signal(x=x)

        if isinstance(x, list) and x.__len__() == 1:
            return self.process_signal(x=x[0])

        else:
            if self.n_processes == 1:
                output = []
                for signal in x:
                    out_tuple = self.process_signal(x=signal)
                    output.append(out_tuple)
                return output
            else:
                with multiprocessing.Pool(self.n_processes) as p:
                    print('It is happening')
                    pfunc = partial(self.process_signal)
                    output = p.map(pfunc, x)
                return output

    def prefilter(self, x):
        x = signal.filtfilt(*self.FILTER_LOW, x, axis=1)
        x = signal.filtfilt(*self.FILTER_HIGH, x, axis=1)
        return x

    def attenuate(self, x):
        if isinstance(self.FILTERS_ATTENUATE, list):
            if self.FILTERS_ATTENUATE.__len__() > 0 and self.ATTENUATE_N > 0:
                for k in range(self.ATTENUATE_N):
                    for f in self.FILTERS_ATTENUATE:
                        x = signal.filtfilt(*f, x, axis=1)
        return x

    def process_signal(self, x):
        """

        Parameters
        ----------
        x


        Returns
        -------

        """
        x = x.copy().squeeze()
        features = []
        msg = []

        segm_size = self._segm_size
        fs = self._fs
        datarate = self.datarate
        sperwelchseg = self._sperwelchseg
        soverlapwelchseg = self._soverlapwelchseg
        fbands = self._fbands
        nfft = self._nfft
        ignore_bands = self._ignore_bands

        if x.shape[0] < fs * segm_size:  # if the signal is shorter than segm_size, appends zeros
            x = np.append(x, np.zeros(int(round(fs * segm_size)) - x.shape[0]))

        xbuffered = buffer(x, fs, segm_size)
        if datarate is True:
            features = features + [1 - (np.isnan(xbuffered).sum(axis=1) / (segm_size * fs))]
            msg = msg + ['DATA_RATE']
        xbuffered = xbuffered - np.nanmean(xbuffered, axis=1).reshape((-1, 1))
        xbuffered[np.isnan(xbuffered)] = 0

        xbuffered = self.prefilter(xbuffered)
        xbuffered = self.attenuate(xbuffered)

        # TODO: REDO FILTERING

        if isinstance(sperwelchseg, type(None)):
            soverlapwelchseg = 0
        else:
            sperwelchseg = int(np.round(sperwelchseg * fs))
            soverlapwelchseg = int(np.round(soverlapwelchseg * fs))
        freq, psd = PSD(xbuffered, fs, nperseg=sperwelchseg, noverlap=soverlapwelchseg, nfft=nfft)
        freq = freq[1:]  # remove 0Hz sample
        psd = psd[:, 1:]

        if ignore_bands.__len__() > 0:
            for eband in ignore_bands:
                psd[:, (freq > eband[0]) & (freq < eband[1])] = np.nan

        inp_params = ObjDict({
            'psd': psd,
            'fs': fs,
            'fbands': fbands,
            'segm_size': segm_size,
            'freq': freq
        })

        for func in self._extraction_functions:
            feature, ftr_name = func(inp_params)
            features = features + feature
            msg = msg + ftr_name
        return features, msg

    @property
    def extraction_functions(self):
        return self._extraction_functions

    @extraction_functions.setter
    def extraction_functions(self, item: list):
        self._extraction_functions = item
        self._verify_extractor_functions()

    @staticmethod
    def _verify_input_fs(item):
        if not isinstance(item, (int, float)):
            raise TypeError('[INPUT TYPE ERROR] Sampling frequency \"fs\" has to be an integer or float!')
        if not item > 0:
            raise ValueError(
                '[INPUT VALUE ERROR] Sampling frequency is required to be higher than 0! Pasted value: ' + str(item))
        return item

    @staticmethod
    def _verify_input_segm_size(item):
        if not isinstance(item, (int, float)):
            raise TypeError(
                '[INPUT TYPE ERROR] A segment size \"segm_size\" is required to be an integer or float. Parsed data type is ' + str(
                    type(item)))
        if not item > 0:
            raise ValueError('[INPUT VALUE ERROR] A segment size \"segm_size\" is required to be  higher than 0!')
        if item == np.inf:
            raise ValueError('[INPUT VALUE ERROR] A segment size \"segm_size\" cannot be Inf')
        return item

    @staticmethod
    def _verify_input_fbands(item):
        if not isinstance(item, (list, np.ndarray)):
            raise TypeError(
                '[INPUT TYPE ERROR] fbands variable has to be of a list or numpy.array type. Pasted value: ' + str(
                    type(item)))
        if not item.__len__() > 0:
            raise ValueError(
                '[INPUT SIZE ERROR] Length of fbands has to be > 0. Current length: ' + str(item.__len__()))
        for idx, subitem in enumerate(item):
            if not subitem.__len__() == 2:
                raise TypeError(
                    '[INPUT SIZE ERROR] Length of each frequency band in fband variable has to contain exactly 2 elements min and max frequency for a given bandwidth. Current size: ' + str(
                        subitem.__len__()))
            if not subitem[0] < subitem[1]:
                raise ValueError('[INPUT VALUE ERROR] For a bandwidth in variable fbands with index ' + str(
                    idx) + ' an error has been found. The first value has to be lower than the second one! Current input: ' + str(
                    subitem))
        return np.array(item)

    @staticmethod
    def _verify_input_x(item):
        if not isinstance(item, (np.ndarray, list)):
            raise TypeError(
                '[INPUT TYPE ERROR] An input signal has to be a type of list or numpy.ndarray. Pasted ' + str(
                    type(item)) + ' instead.')

        if isinstance(item, np.ndarray):
            if not (item.shape.__len__() == 1 or item.shape.__len__() == 2):
                raise TypeError(
                    '[INPUT SIZE ERROR] An input signal has to consist of an input of a single dimension for a single signal, 2D numpy.ndarray field for multiple signals (n_signal, signal_length), or list containing multiple fields with a single signal in each of these cells.')

        if isinstance(item, list):
            for subitem in item:
                if not isinstance(subitem, np.ndarray):
                    raise TypeError(
                        '[INPUT SIZE ERROR] An input signal has to consist of an input of a single dimension for a single signal, 2D numpy.ndarray field for multiple signals (n_signal, signal_length), or list containing multiple fields with a single signal in each of these cells.')

        return item

    @staticmethod
    def _verify_input_n_processes(item):
        if not isinstance(item, int):
            raise TypeError('[INPUT TYPE ERROR] Input n_processes has to be of a type int. Type ' + str(
                type(input)) + ' has found instead.')
        if item < 1:
            raise ValueError(
                '[INPUT VALUE ERROR] Number of processes dedicated to feature extraction should be > than 0.')
        if item > multiprocessing.cpu_count() / 2:
            raise PendingDeprecationWarning(
                '[INPUT VALUE ERROR] Number of processes dedicated to feature extraction shouldn\'t be higher than half of the number of processors. This can significantly slow down the processing time and decrease performance. Value is decreased to a number ' + str(
                    multiprocessing.cpu_count() / 2))
            return int(multiprocessing.cpu_count() / 2)
        return item

    def _verify_extractor_functions(self):
        if self._extraction_functions.__len__() < 1:
            raise TypeError('')

        for idx, func in enumerate(self._extraction_functions):
            if not callable(func):
                raise TypeError('[FUNCTION ERROR] A feature extraction function ' + str(func) + ' with an index ' + str(
                    idx) + ' is not callable')

























# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
Tools for handling features and feature labels during classification, data preparation and evaluation
"""

import numpy as np
import pandas as pd
from sklearn.neighbors import LocalOutlierFactor
from sklearn.metrics import f1_score, precision_score, recall_score, confusion_matrix, classification_report, cohen_kappa_score, accuracy_score, precision_recall_curve, auc


def zscore(x):
    """
    Calculates Z-score
    Parameters
    ----------
    x : numpy ndarray
        shape[n_samples, n_features]

    Returns
    ----------
    normalized_features : numpy ndarray
    """

    return (x - x.mean(axis=0).reshape(1, -1)) / x.std(axis=0).reshape(1, -1)


def find_category_outliers(x, y=None):
    """
    Finds outliers for each category within data.
    Check website: https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.LocalOutlierFactor.html

    Parameters
    ----------
    x : numpy ndarray
        shape[n_samples, n_features]
    y : list or numpy array
        string or int indexes for each category

    Returns
    -------
    list
        position index list with detected outliers

    """
    if isinstance(y, type(None)):
        y = np.zeros(x.shape[0])


    ycat = np.unique(y)
    to_del = []

    #x_temp = x.copy()
    #for k in range(x_temp.shape[1]):
    #    to_del = to_del + list(np.where(np.abs(x_temp[:, k]) > 4)[0])
    x = zscore(x)

    for yc in ycat:
        positions = np.where(np.array(y) == yc)[0]
        x_sub = x[positions]
        pred = LocalOutlierFactor().fit_predict(x_sub)
        to_del = to_del + list(positions[np.where(pred == -1)[0]])
    return to_del


def print_classification_scores(Y, YY, N_merge=False):
    """
    Prints classification report for sleep scoring labels.
    """

    labels = ['AWAKE', 'N2', 'N3', 'REM']
    kappa = cohen_kappa_score(Y, YY)
    f1 = f1_score(Y, YY, average='weighted')
    acc = accuracy_score(Y, YY)
    un_rt_YY = (YY == 'UNKNOWN').sum() / YY.shape[0]
    un_rt_Y = (Y == 'UNKNOWN').sum() / Y.shape[0]
    cmat = confusion_matrix(Y, YY, labels=labels)


    Y = Y[YY != 'UNKNOWN']
    YY = YY[YY != 'UNKNOWN']

    YY = YY[Y != 'UNKNOWN']
    Y = Y[Y != 'UNKNOWN']


    print('Confusion Matrix:')
    print(cmat)


    print('###########')
    print('All stages')
    print('Kappa-score: {:.2f}'.format(kappa), ' F1: {:.2f}'.format(f1), ' Accuracy: {:.2f}'.format(acc), ' Unknown rate gold/clf: {:.4f} / {:.4f}'.format(un_rt_Y, un_rt_YY))
    print(classification_report(Y, YY))

    for state in labels:
        Y_ = Y.copy()
        YY_ = YY.copy()
        Y_[Y_ != state] = 'all'
        YY_[YY_ != state] = 'all'
        print(state, ' vs. all - kappa: {:.2f}'.format(cohen_kappa_score(Y_, YY_)), ' F1: {:.2f}'.format(f1_score(Y_==state, YY_==state, average='micro')), ' Accuracy: {:.2f}'.format(accuracy_score(Y_, YY_)))

    Y_ = Y.copy()
    YY_ = YY.copy()
    Y_ = replace_annotations(Y_, old_key=['N2', 'N3'], new_key='N')
    YY_ = replace_annotations(YY_, old_key=['N2', 'N3'], new_key='N')
    kappa_Nmerged = cohen_kappa_score(Y_, YY_)

    state = 'N'
    Y_[Y_ != state] = 'all'
    YY_[YY_ != state] = 'all'
    print(state, ' vs. all - kappa: {:.2f}'.format(cohen_kappa_score(Y_, YY_)), ' F1: {:.2f}'.format(f1_score(Y_=='N', YY_=='N', average='micro')), ' Accuracy: {:.2f}'.format(accuracy_score(Y_, YY_)))
    print('Kappa N, merged {:.2f}'.format(kappa_Nmerged))

    if N_merge:


        kappa = cohen_kappa_score(Y, YY)
        f1 = f1_score(Y, YY, average='weighted')
        acc = accuracy_score(Y, YY)
        un_rt_YY = (YY == 'UNKNOWN').sum() / YY.shape[0]
        un_rt_Y = (Y == 'UNKNOWN').sum() / Y.shape[0]
        cmat = confusion_matrix(Y, YY, labels=labels)
        labels = ['AWAKE', 'N', 'REM']

        print('###########')
        print('N - merged')
        print('Kappa-score: {:.2f}'.format(kappa), ' F1: {:.2f}'.format(f1), ' Accuracy: {:.2f}'.format(acc), ' Unknown rate gold/clf: {:.4f} / {:.4f}'.format(un_rt_Y, un_rt_YY))
        print(classification_report(Y, YY))

        for state in labels:
            print(state, ' vs. all - kappa: {:.2f}'.format(cohen_kappa_score(Y_, YY_)), ' F1: {:.2f}'.format(f1_score(Y_==state, YY_==state, average='binary')), ' Accuracy: {:.2f}'.format(accuracy_score(Y_, YY_)))


def get_classification_scores(Y, YY, labels=None):
    """
    
    Returns a classification report. All values are already in a formated string.

    """
    if isinstance(labels, type(None)): labels = np.unique(list(Y) + list(YY))
    un_rt_YY = (YY == 'UNKNOWN').sum() / YY.shape[0]
    un_rt_Y = (Y == 'UNKNOWN').sum() / Y.shape[0]

    Y = Y[YY != 'UNKNOWN']
    YY = YY[YY != 'UNKNOWN']

    YY = YY[Y != 'UNKNOWN']
    Y = Y[Y != 'UNKNOWN']


    kappa = cohen_kappa_score(Y, YY)
    f1 = f1_score(Y, YY, average='weighted')
    acc = accuracy_score(Y, YY)

    score = {}
    score['kappa_all'] = '{:.3f}'.format(kappa)
    score['f1_all']  = '{:.3f}'.format(f1)
    score['accuracy_all']  = '{:.3f}'.format(acc)
    score['unknown'] = '{:.3f}'.format(un_rt_YY)


    for state in labels:
        Y_ = Y.copy()
        YY_ = YY.copy()
        Y_[Y_ != state] = 'all'
        YY_[YY_ != state] = 'all'
        score['kappa_'+state] = '{:.3f}'.format(cohen_kappa_score(Y_, YY_))
        score['f1_'+state]  = '{:.3f}'.format(f1_score(Y_==state, YY_==state, average='binary'))
        score['accuracy_'+state]  = '{:.3f}'.format(accuracy_score(Y_, YY_))

    Y_ = replace_annotations(Y, old_key=['N2', 'N3'], new_key='N')
    YY_ = replace_annotations(YY, old_key=['N2', 'N3'], new_key='N')
    kappa_Nmerged = cohen_kappa_score(Y_, YY_)

    state = 'N'
    score['kappa_'+state] = '{:.3f}'.format(cohen_kappa_score(Y_, YY_))
    score['f1_'+state]  = '{:.3f}'.format(f1_score(Y_=='N', YY_=='N', average='weighted'))
    score['accuracy_'+state]  = '{:.3f}'.format(accuracy_score(Y_, YY_))
    return score


def augment_features(x, feature_names=None, feature_indexes=[], operation=None, mutual=False, operation_str = ''):
    """
    Augments features with entered operations (mutual - between features such as ``*``, ``/``, ``+``, ``-``, ....; non mutual - log, exp, power, ...)

    Parameters
    ----------
    x : numpy ndarray
            shape[n_samples, n_features]
    feature_names : list or numpy array of strings, optional
        names of features
    feature_indexes : list or numpy array
        indexes of features which will be augmented
    operation : function
        callable function which will be applied on existing features.
    mutual :  bool
        indicates whether operation is applied on single feature e.g. np.log10, or on 2 parameters e.g. np.divide
        if mutual = True, then applied on all feature combination specified in feature_indexes

    Returns
    -------
    numpy ndarray -> shape[n_samples, n_features]

    """
    if isinstance(feature_indexes, type(None)):
        feature_indexes = np.arange(x.shape[1])

    if not isinstance(feature_names, type(None)):
        feature_names = list(feature_names)

    # If mutual is true - augments single features
    if mutual is False:
        for idx, ftr_idx in enumerate(feature_indexes):
            temp_x = operation(x[:, ftr_idx])
            x = np.concatenate((x, temp_x.reshape(-1, 1)), axis=1)
            if not isinstance(feature_names, type(None)):
                feature_names = feature_names + [operation_str + feature_names[ftr_idx]]

    # augments feature combination
    else:
        for idx_1, ftr_idx_1 in enumerate(feature_indexes):
            feature_sub_indexes = feature_indexes[idx_1:]
            if not isinstance(feature_names, type(None)):
                feature1_name = feature_names[ftr_idx_1]

            for idx_2, ftr_idx_2 in enumerate(feature_sub_indexes):
                if ftr_idx_1 != ftr_idx_2:
                    temp_x = operation(x[:, ftr_idx_1], x[:, ftr_idx_2])
                    x = np.concatenate((x, temp_x.reshape(-1, 1)), axis=1)

                    if not isinstance(feature_names, type(None)):
                        feature2_name = feature_names[ftr_idx_2]
                        feature_names = feature_names + [feature1_name + ' ' + operation_str + ' ' + feature2_name]

    if not isinstance(feature_names, type(None)):
        return x, feature_names

    return x


def remove_features(x, feature_names=None, to_del=None):
    """
    Removes features

    Parameters
    ----------
    x : numpy ndarray
        shape[n_samples, n_features]
    feature_names : list or numpy array, optional
        names of features
    to_del :

    """
    x = np.delete(x, to_del, 1)
    if not isinstance(feature_names, type(None)):
        feature_names = np.delete(feature_names, to_del, 0)
        return x, feature_names
    return x


def remove_samples(x, y=None, to_del=None):
    """
    Removes samples

    Parameters
    ----------
    x : numpy ndarray / list / pd.DataFrame
        shape[n_samples, n_features]
    y : list or numpy array, optional
        category reference for each sample
    to_del :

    """
    to_del = np.array(to_del)
    if to_del.dtype == np.bool: # if to_del is array of bools
        if x.__len__() != to_del.__len__():
            raise AssertionError('If to_del is a bool array, must be the same length as x')
        to_del = np.where(to_del)[0]


    if isinstance(x, (np.ndarray, list, pd.DataFrame)): # if x is array of parameters / list - can process also reference y
        if isinstance(x, np.ndarray):
            x = np.delete(x, to_del, 0)
        if isinstance(x, list):
            x = [x_ for idx, x_ in enumerate(x) if idx not in to_del]
        if isinstance(x, pd.DataFrame): # if dataframe
            x = x.drop(to_del, axis=0).reset_index(drop=True)

        if not isinstance(y, type(None)):
            y = np.delete(np.array(y), to_del, 0)
            return x, y
        return x


def balance_classes(x, y, std_factor=0.0):
    """
    Balances unbalanced classes in dataset by extending the sample array with same samples, possibly with introduced
    noise. Detects classes from y variable and number of samples per category. Duplicates samples from the categories
    with lower number of samples. std_factor gives the level of noise introduced into duplicated samples relatively to
    the std of a given dimension for a given category.

    Parameters
    ----------
    x : numpy ndarray
        shape[n_samples, n_features]
    y : list or numpy array
        string or int indexes for each category
    std_factor : float
        Amount of noise introduced into duplicated features relatively to std of a given feature within a category.

    Returns
    -------
    numpy ndarray
        x - samples
    list
        y - categories
    """

    # Data augmentation
    cat_members = np.array([(y == c).sum() for c in np.unique(y)])
    for idx, cat in enumerate(np.unique(y)):
        num = (y == cat).sum()
        target_num = cat_members.max()
        generate = target_num - num
        if generate > 0:
            src_idxes = np.array(list(np.where(y == cat)[0]) * int(np.ceil(generate / num)))
            x_aug = x[src_idxes, :]
            x_aug = x_aug + (np.random.randn(x_aug.shape[0], x_aug.shape[1]) * x_aug.std(axis=0) * std_factor)
            x_aug = x_aug[:generate]
            x = np.concatenate((x,  x_aug[:generate]), axis=0)
            y = np.array(list(y) + [cat] * generate)
    return x, y


def replace_annotations(Y, old_key=None, new_key=None):
    """
    Replaces annotation names in a numpy array or list
    """
    Y = list(Y)
    if not isinstance(old_key, list):
        old_key = [old_key]
    Y_new = []
    for Y_ in Y:
        if Y_ in old_key:
            Y_new.append(new_key)
        else:
            Y_new.append(Y_)
    return np.array(Y_new)





# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from copy import deepcopy

def _check_dimensions(x, n_low=1, n_high=3):
    if x.shape[1] > n_high or x.shape[1] < n_low:
        raise AssertionError('[ERROR]: Data has to be 1-3 dimensions')
    return True


def _check_scale(x, m):
    if x.shape[1] != m.__len__():
        raise AssertionError('[ERROR]: N-dim of data has to be the same as scale coefficients')
    return True


def translate(x, m):
    _check_scale(x, m)
    for idx, s in enumerate(m):
        x[:, idx] = x[:, idx] + s
    return x


def scale(x, m):
    _check_scale(x, m)
    sc = np.array(m).reshape(1, -1)
    mn = x.mean(axis=0).reshape(1, -1)
    x = (x-mn)*sc + mn
    return x


def rotate(x, angl):
    x = deepcopy(x)
    if isinstance(angl, (tuple, np.ndarray, list)):
        if angl.__len__() == 3 and x.shape[1] == 3:
            return _rotate_3d(x, angl)
        if angl.__len__() == 1:
            angl = angl[0]
    return _rotate_2d(x, angl)


def _rotate_2d(x, angl):
    m = get_rot_2d(angl)
    mn = x.mean(axis=0).reshape(1, -1)
    return np.dot(x-mn, m) + mn


def get_rot_2d(an):
    an = 2*np.pi*an/360
    m = np.array([
        [np.cos(an), -np.sin(an)],
        [np.sin(an), np.cos(an)]
    ])
    return m


def _rotate_3d(x, angl):
    ms = get_rot_3d(angl)
    mn = x.mean(axis=0).reshape(1, -1)
    x = x - mn
    for m in ms:
        x = np.dot(x, m)
    x = x + mn
    return x


def get_rot_3d(an):
    return [
        get_rot_3d_x(an[0]),
        get_rot_3d_y(an[1]),
        get_rot_3d_z(an[2])
    ]


def get_rot_3d_x(an):
    an = 2*np.pi*an/360
    m = np.array([
        [1, 0, 0],
        [0, np.cos(an), -np.sin(an)],
        [0, np.sin(an), np.cos(an)]
    ])
    return m


def get_rot_3d_y(an):
    an = 2*np.pi*an/360
    m = np.array([
        [np.cos(an), 0, np.sin(an)],
        [0, 1, 0],
        [-np.sin(an), 0,  np.cos(an)]
    ])
    return m


def get_rot_3d_z(an):
    an = 2*np.pi*an/360
    m = np.array([
        [np.cos(an), -np.sin(an), 0],
        [np.sin(an), np.cos(an), 0],
        [0, 0, 1]
    ])
    return m


def get_mutual_vectors(x, y=None):
    leg = []
    v = []
    for idx, x_ in enumerate(x):
        temp_x = x_ - x
        if not isinstance(y, type(None)):
            temp_y = np.array([y[idx] + '-' + y_ for y_ in y])
        v += [temp_x]
        leg += [temp_y]
    v = np.concatenate(v, axis=0)

    if not isinstance(y, type(None)):
        leg = np.concatenate(leg)
        return v, leg
    return v








# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from datetime import timedelta
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from datetime import timedelta, datetime


def plot_hypnogram(orig_df, hypnogram_values=None, hypnogram_colors=None, fontsize=12, fig=None, night_start=22):
    """
    Creates a Matplotlib figure of spectrogram from the hypnogram. Time must be in a time-zone aware format.

    Parameters
    ----------
    orig_df : hypnogram
    hypnogram_values : dict
        dict of a y-axis values for each sleep state
    hypnogram_colors : dict
        dict of color hex codes for each sleep state
    fontsize : int
        Fontsize
    fig : figure
        Already existing figure object.
    night_start : int
        A hour when does the night begin

    Returns
    -------

    """
    _hypnogram_values = {
        'AWAKE': 6,
        'Arousal': 5,
        'SLP': 4.5,
        'REM': 4,
        'N1': 3,
        'N2': 2,
        'N3': 1,
    }

    _hypnogram_colors = {
        'AWAKE': '#e7b233',
        'Arousal': '#d44b05',
        'SLP': '#3500d3',
        'REM': '#3500d3',
        'N1': '#2bc7c4',  # 2b7cc7
        'N2': '#2b5dc7',
        'N3': '#000000',
    }

    if isinstance(hypnogram_colors, type(None)):
        hypnogram_colors = _hypnogram_colors

    if isinstance(hypnogram_values, type(None)):
        hypnogram_values = _hypnogram_values

    def set_hypnogram_properties(x, ref_dict):
        return ref_dict[x.annotation]

    orig_df['state_id'] = orig_df.apply(lambda x: set_hypnogram_properties(x, hypnogram_values), axis=1)
    orig_df['state_color'] = orig_df.apply(lambda x: set_hypnogram_properties(x, hypnogram_colors), axis=1)
    df_arrousals = orig_df.loc[orig_df.annotation == 'Arrousal'].reset_index(drop=True)
    df = orig_df.loc[orig_df.annotation != 'Arrousal'].reset_index(drop=True)
    new_df = pd.DataFrame()
    for idx, row in enumerate(df.iterrows()):  # if 2 cons. states are same, merges them
        appbl = True
        if idx > 0:
            if new_df.iloc[-1].state_id == row[1].state_id and new_df.iloc[-1].end == row[1].start:
                appbl = False

        if appbl == True:
            new_df = new_df.append(row[1], ignore_index=True)
        else:
            new_df.loc[new_df.__len__() - 1, 'end'] = row[1].end
    df = new_df

    x_start = np.array(df['start'])
    x_end = np.array(df['end'])
    try:
        for k, time_sample in enumerate(x_start): x_start[k] = time_sample.to_pydatetime()
        for k, time_sample in enumerate(x_end): x_end[k] = time_sample.to_pydatetime()
    except:
        for k, time_sample in enumerate(x_start): x_start[k] = time_sample
        for k, time_sample in enumerate(x_end): x_end[k] = time_sample

    if not fig:
        plt.figure(dpi=200)
    plt.xlim(x_start[0], x_end[-1])
    """
    # set background color for days
    for idx, day_id in enumerate(np.unique(df.day)):
        if idx % 2 == 0:
            background_color = 'gray'
            background_alpha = 0.1
        else:
            background_color = 'gray'
            background_alpha = 0.3

        day_start = x_start[df.day == day_id][0]
        day_end = x_end[df.day == day_id][-1]
        plt.axvspan(day_start, day_end, facecolor=background_color, alpha=background_alpha)
    """

    # set background color for nights
    for idx, day_id in enumerate(np.unique(df.day)):
        background_color = 'gray'
        background_alpha = 0.3
        day_start = x_start[df.day == day_id][0]
        night_start_ = datetime(
            year=day_start.year,
            month=day_start.month,
            day=day_start.day,
            hour=night_start,
            tzinfo=day_start.tzinfo
        )
        night_end_ = night_start_ + timedelta(hours=12)
        plt.axvspan(night_start_, night_end_, facecolor=background_color, alpha=background_alpha)

    # plot columns
    for idx, row in enumerate(df.iterrows()):
        val = row[1]['state_id']
        clr = row[1]['state_color']

        plt.fill_between(
            [x_start[idx], x_end[idx]],
            [val, val],
            color=clr,
            alpha=0.5,
            linewidth=0
        )

    for idx in range(df.__len__() - 1):
        val0 = df.state_id[idx]
        val1 = df.state_id[idx + 1]
        start0 = df.start[idx]
        start1 = df.start[idx + 1]
        end0 = df.end[idx]
        end1 = df.end[idx + 1]

        if val0 == val1:
            if end0 == start1:
                x = [start0, start1]
                y = [val0, val1]
            else:
                x = [start0, end0]
                y = [val0, val0]
        else:
            if end0 == start1:
                x = [start0, end0, start1]
                y = [val0, val0, val1]
            else:
                x = [start0, end0]
                y = [val0, val0]

        plt.plot(x, y, color='black', alpha=1, linewidth=1)

    x = [start1, end1]
    y = [val1, val1]
    plt.plot(x, y, color='black', alpha=1, linewidth=1)

    # plot arrousals
    for row in df_arrousals.iterrows():
        val = row[1].state_id
        clr = row[1].state_color
        plt.fill_between(
            # plt.plot(
            [row[1].start.to_pydatetime(), row[1].start.to_pydatetime(row[1].start), row[1].end.to_pydatetime(),
             row[1].end.to_pydatetime(row[1].end)],
            [0, val, val, 0],
            color=clr,
            alpha=1,
            linewidth=1
        )

    # format y ticks
    plt.yticks(list(hypnogram_values.values()), hypnogram_values.keys())
    for ticklabel in plt.gca().get_yticklabels():
        clr = hypnogram_colors[ticklabel._text]
        ticklabel.set_color(clr)
        # ticklabel.set_fontsize(fontsize)

    # plot y grid
    for idx, key in enumerate(hypnogram_values.keys()):
        clr = hypnogram_colors[key]
        val = hypnogram_values[key]
        plt.plot([x_start[0], x_end[-1]], [val, val], color=clr, linewidth=0.7, alpha=0.7, linestyle=':')

    # format x_ticks
    plt.gcf().autofmt_xdate()
    # formatter = mdates.DateFormatter("%H:%M") #mdates.DateFormatter("%H:%M", tz=tz.tzlocal())
    formatter = mdates.DateFormatter("%H:%M", tz=df.start[0].tzinfo)
    plt.gcf().get_axes()[0].xaxis.set_major_formatter(formatter)

    # plot hour x grid
    plt.grid(True, axis='x', alpha=1, linewidth=0.5, linestyle=':')

    # axes labels
    # plt.title('Days  ' + df.start[0].strftime('%d.%m') +'-' + df.iloc[-1].end.strftime('%d.%m'))
    plt.xlabel('\n Time [' + df.start[0].strftime('%d.%m.%Y') + ' - ' + df.iloc[-1].end.strftime('%d.%m.%Y') + ']',
               fontsize=fontsize)
    plt.ylabel('Sleep state', fontsize=fontsize)
    plt.gca().tick_params(axis='both', which='major', labelsize=fontsize)




# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from io import StringIO
import xml.etree.ElementTree as ET
import os
import pandas as pd
from dateutil import tz
from datetime import datetime
import os
import re
from datetime import datetime
import pytz
from copy import deepcopy
import argparse
import xml.etree.ElementTree as ET
from _annotation_tools.XML_CyberPSG import *
import uuid

xsi =  "http://www.w3.org/2001/XMLSchema-instance"
xmlns =  "http://tempuri.org/CyberPSG.xsd"
ns = {"xmlns:xsi": xsi}

class TwoWayDict(dict):
    def __init__(self, d=None):
        if not isinstance(d, type(None)):
            if isinstance(d, dict):
                for k, v in d.items():
                    self[k] = v



    def __setitem__(self, key, value):
        # Remove any previous connections with these values
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        dict.__setitem__(self, key, value)
        dict.__setitem__(self, value, key)

    def __delitem__(self, key):
        dict.__delitem__(self, self[key])
        dict.__delitem__(self, key)

    def __len__(self):
        return dict.__len__(self) // 2

class CyberPSGFile:
    def __init__(self, path=None):
        self.namespaces = TwoWayDict() # list of namespaces - key=prefix; value=uri
        self.Element = None
        self.path = path
        self.strp_format = '%Y-%m-%dT%H:%M:%S.%f'

        if path:
            if os.path.isfile(path):
                self.read_file(path_xml=path)


    def register_namespace(self, prefix, uri):
        if not prefix in self.namespaces.keys():
            self.namespaces[prefix] = uri
            ET.register_namespace(prefix, uri)

    def parse_namespaces(self, path_xml):
        schema = open(path_xml, 'r').read()
        namespaces = dict([
            node for _, node in ET.iterparse(StringIO(schema), events=['start-ns'])
        ])
        for k, v in namespaces.items(): self.register_namespace(k, v)

    def register_CyberPSG_namespace(self):
        self.register_namespace('', "http://tempuri.org/CyberPSG.xsd")
        self.register_namespace('xsi', "http://www.w3.org/2001/XMLSchema-instance")


    def read_file(self, path_xml):
        self.parse_namespaces(path_xml)
        self.Element = ET.parse(path_xml).getroot()

    def get_annotation_types(self):
        types = {}
        for Element in self.Element:
            tag = Element.tag
            if '}' in tag: tag = tag.split('}')[-1]
            if tag == 'AnnotationTypes':
                for AnnotationTypeElement in Element:
                    name = None
                    id = None
                    electrode = None
                    for SubElement in AnnotationTypeElement:
                        subtag = SubElement.tag
                        if '}' in subtag: subtag = subtag.split('}')[-1]
                        if subtag == 'id': id = SubElement.text
                        if subtag == 'name': name = SubElement.text
                    types[id] = name
        return types


    def get_annotations(self):
        types = self.get_annotation_types()
        types = TwoWayDict(types)
        annotations = []

        for Element in self.Element:
            tag = Element.tag
            if '}' in tag: tag = tag.split('}')[-1]
            if tag == 'Annotations':
                for AnnotationElement in Element:
                    annot = {
                        'annotation': None,
                        'startTimeUtc': None,
                        'endTimeUtc': None
                    }

                    for SubElement in AnnotationElement:
                        subtag = SubElement.tag
                        if '}' in subtag: subtag = subtag.split('}')[-1]
                        if subtag == 'annotationTypeId':  annot['annotationTypeId'] = types[types[SubElement.text]]
                        if subtag == 'startTimeUtc':  annot['startTimeUtc'] = SubElement.text
                        if subtag == 'endTimeUtc':  annot['endTimeUtc'] = SubElement.text
                        if subtag == 'channelName': annot['channel'] = SubElement.text
                    annotations += [annot]
        return annotations

    def get_hypnogram(self):
        types = self.get_annotation_types()
        types = TwoWayDict(types)
        annotations = []

        for Element in self.Element:
            tag = Element.tag
            if '}' in tag: tag = tag.split('}')[-1]
            if tag == 'Annotations':
                for AnnotationElement in Element:
                    annot = {
                        'annotation': None,
                        'start': None,
                        'end': None
                    }

                    for SubElement in AnnotationElement:
                        subtag = SubElement.tag
                        if '}' in subtag: subtag = subtag.split('}')[-1]
                        if subtag == 'annotationTypeId':  annot['annotation'] = types[SubElement.text]
                        if subtag == 'startTimeUtc':
                            s = SubElement.text
                            if '.' in s:
                                frmt = self.strp_format
                            else:
                                frmt = self.strp_format.split('.')[0]
                            utc = datetime.strptime(s, frmt)
                            utc = utc.replace(tzinfo=tz.tzutc())
                            annot['start'] = utc
                        if subtag == 'endTimeUtc':
                            s = SubElement.text
                            if '.' in s:
                                frmt = self.strp_format
                            else:
                                frmt = self.strp_format.split('.')[0]
                            utc = datetime.strptime(s, frmt)
                            utc = utc.replace(tzinfo=tz.tzutc())
                            annot['end'] = utc
                        if subtag == 'channelName': annot['channel'] = SubElement.text
                    annotations += [annot]
        return annotations

class CyberPSG_XML_Writter:
    __version__ = '1.0.1'
    """
    Version 1.0.0 update
    Database of UUIDs implemented
    
    Version 1.01 update
    adding channel name to the export
    """


    _xsi =  "http://www.w3.org/2001/XMLSchema-instance"
    _xmlns =  "http://tempuri.org/CyberPSG.xsd"
    _ns = {"xmlns:xsi": xsi}

    standard_UUID = {
        'AWAKE_best': '00000000-898a-4b80-8ed8-000000000000',
        'N1_best' : '00000000-898a-4b80-8ed8-111111111111',
        'N2_best': '00000000-898a-4b80-8ed8-222222222222',
        'N3_best': '00000000-898a-4b80-8ed8-333333333333',
        'REM_best': '00000000-898a-4b80-8ed8-444444444444',
        'UNKNOWN_best': '00000000-898a-4b80-8ed8-555555555555',
        'Arousal_best': '00000000-898a-4b80-8ed8-666666666666',
        'N_best': '00000000-898a-4b80-8ed8-777777777777',
        'SLP_best': '00000000-898a-4b80-8ed8-888888888888',
        'Sleep stage N1': '00000001-898a-4b80-8ed8-000000000001',
        'Sleep stage N2': '00000001-898a-4b80-8ed8-000000000002',
        'Sleep stage N3': '00000001-898a-4b80-8ed8-000000000003',
        'Sleep stage R':  '00000001-898a-4b80-8ed8-000000000004',
        'Sleep stage W':  '00000001-898a-4b80-8ed8-000000000005',
        'IED':  '00000001-898a-4b80-8ed8-000000000010',
        'IED_best':  '00000001-898a-4b80-8ed8-000000000011',
    }




    annotation_keys = [k[:-5] for k in standard_UUID.keys()]


    def __init__(self, path):
        for attr, uri in self._ns.items():
            ET.register_namespace(attr.split(":")[1], uri)

        self.AnnotationData = myXML_AnnotationData.init_EmptyInstance(file_path=path)
        self.AnnotationData.attributes['xmlns'] = self._xmlns

        self.AnnotationGroups = self.AnnotationData.add_child(myXML_AnnotationGroups)
        self.AnnotationTypes = self.AnnotationData.add_child(myXML_AnnotationTypes)
        self.Annotations = self.AnnotationData.add_child(myXML_Annotations)

    @property
    def nAnnotationGroups(self):
        return self.AnnotationGroupKeys.__len__()

    @property
    def AnnotationGroupKeys(self):
        return [AnnotationGroup['name'][0].param for AnnotationGroup in self.AnnotationGroups['AnnotationGroup']]

    @property
    def AnnotationGroupIDs(self):
        return [AnnotationGroup['id'][0].param for AnnotationGroup in self.AnnotationGroups['AnnotationGroup']]

    def add_AnnotationGroup(self, name_string='', uuid_=None):
        if name_string.__len__() == 0:
            name_string = 'AnnotationGroup{0}'.format(self.AnnotationGroups['AnnotationGroup'].__len__())

        for AnnotationGroup in self.AnnotationGroups['AnnotationGroup']:
            if AnnotationGroup['name'] == name_string:
                raise KeyError("AnnotationGroup \"" + name_string + "\" already exists!")

        AnnotationGroup = self.AnnotationGroups.add_child(myXML_AnnotationGroup)
        ID = AnnotationGroup.add_child(myXML_Id)
        if not uuid_:
            ID.param = str(uuid.uuid4())
        else:
            ID.param = str(uuid_)
        name = AnnotationGroup.add_child(myXML_Name)
        name.param = name_string

    def remove_AnnotationGroup(self, item):
        if not (isinstance(item, int) and item >= 0 and item < self.AnnotationGroups['AnnotationGroup'].__len__()) and not isinstance(item, str):
            raise KeyError('Annotation group identifier must be an integer indexing Annotation groups or string representing a category name.')

        for idx, AnnotationGroup in enumerate(self.AnnotationGroups['AnnotationGroup']):
            if AnnotationGroup['name'][0].param == item:
                item = idx

        self.AnnotationGroups['AnnotationGroup'][item].Element.clear()
        del self.AnnotationGroups.AnnotationGroup[item]

    @property
    def nAnnotationTypes(self):
        return self.AnnotationTypeKeys.__len__()

    @property
    def AnnotationTypeKeys(self):
        return [AnnotationType['name'][0].param for AnnotationType in self.AnnotationTypes.AnnotationType]

    @property
    def AnnotationTypeIDs(self):
        return [AnnotationType['id'][0].param for AnnotationType in self.AnnotationTypes.AnnotationType]

    def add_AnnotationType(self, name_string="", groupAssociationId=None, color='#FF6C1FBA', note="", startsWithEpoch=False, stdDurationInSec=0):
        if name_string.__len__() == 0:
            name_string = 'AnnotationType{0}'.format(self.AnnotationTypes.__len__())

        # if name_string+'_aisc' in list(self.standard_UUID.keys()):
        #     name_string += '_aisc'
        if name_string+'_best' in list(self.standard_UUID.keys()):
            name_string += '_best'

        #print(name_string)
        #print(list(self.standard_UUID.keys()))
        #print(name_string+'_aisc' in list(self.standard_UUID.keys()))

        for AnnotationType in self.AnnotationTypes.AnnotationType:
            if AnnotationType.name[0].param == name_string:
                raise KeyError("AnnotationType \"" + name_string + "\" already exists!")

        AnnotationType = self.AnnotationTypes.add_child(myXML_AnnotationType)

        if name_string in list(self.standard_UUID.keys()):
            uuid_key = self.standard_UUID[name_string]
        else:
            uuid_key = str(uuid.uuid4())

        #name_string = name_string + '_aisc'
        ID = AnnotationType.add_child(myXML_Id, param=uuid_key)
        name = AnnotationType.add_child(myXML_Name, param=name_string)
        color = AnnotationType.add_child(myXML_Color, param=color)
        if note.__len__() > 0:
            note = AnnotationType.add_child(myXML_Note, param=note)

        if isinstance(groupAssociationId, str):
            is_AGroup_set = False
            for idx, AGroupName in enumerate(self.AnnotationGroupKeys):
                if AGroupName == groupAssociationId:
                    groupAssociationId = idx
                    is_AGroup_set = True

            if is_AGroup_set is False:
                raise KeyError('AnnotationType cannot be assigned to an AnnotationGroup \"' + groupAssociationId + '\". Available AnnotationGroups \"' + str(self.AnnotationGroupKeys ))

        if isinstance(groupAssociationId, int):
            group_id = self.AnnotationGroups.AnnotationGroup[groupAssociationId].id[0].param
            groupAssociation = AnnotationType.add_child(myXML_GroupAsociations)
            gAID = groupAssociation.add_child(myXML_Id, param=group_id)

    def remove_AnnotationType(self, item):
        if not (isinstance(item, int) and item >= 0 and item < self.AnnotationTypes.AnnotationType.__len__()) and not isinstance(item, str):
            raise KeyError('Annotation type identifier must be an integer indexing Annotation types or string representing a category name.')

        for idx, AnnotationTypeKey in enumerate(self.AnnotationTypesKeys):
            if AnnotationTypeKey == item:
                item = idx

        self.AnnotationTypes.AnnotationType[item].Element.clear()
        del self.AnnotationTypes.AnnotationType[item]

    @property
    def nAnnotations(self):
        return self.Annotations.Annotation.__len__()

    @property
    def AnnotationIDs(self):
        return [Annotation['id'][0].param for Annotation in self.Annotations.Annotation]

    def add_Annotation(self, start_datetime:datetime, end_datetime:datetime, AnnotationTypeId=0, note='', type='GlobalAnnotation', channelName=''):
        # if AnnotationTypeId + '_aisc' in list(self.standard_UUID.keys()):
        #     AnnotationTypeId += '_aisc'

        if AnnotationTypeId + '_best' in list(self.standard_UUID.keys()):
            AnnotationTypeId += '_best'

        start_s = start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")
        end_s = end_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")

        Annotation = self.Annotations.add_child(myXML_Annotation)
        Annotation.attributes[ET.QName(xsi, "type")] = type
        ID = Annotation.add_child(myXML_Id, param=str(uuid.uuid4()))

        startTimeUtc = Annotation.add_child(myXML_StartTimeUtc, param=start_s)
        endTimeUtc = Annotation.add_child(myXML_endTimeUtc, param=end_s)

        if isinstance(AnnotationTypeId, str):
            is_AType_set = False
            for idx, ATypepName in enumerate(self.AnnotationTypeKeys):
                if ATypepName == AnnotationTypeId:
                    AnnotationTypeId = idx
                    is_AType_set = True

            if is_AType_set is False:
                raise KeyError('AnnotationType cannot be assigned to an AnnotationGroup \"' + AnnotationTypeId + '\". Available AnnotationGroups \"' + str(self.AnnotationGroupKeys ))

        ATypeId_s = self.AnnotationTypeIDs[AnnotationTypeId]
        AnnotationTypeId = Annotation.add_child(myXML_AnnotationTypeId, param=ATypeId_s)

        if channelName:
            chName = Annotation.add_child(myXML_ChannelName, param=channelName)
            vertPosition = Annotation.add_child(myXML_VerticalPositionPercentage, param=0.25)

        if note.__len__() > 0:
            note = Annotation.add_child(myXML_Note, param=note)

    def remove_Annotation(self, item):
        if not (isinstance(item, int) and item >= 0 and item < self.Annotations.Annotation.__len__()) and not isinstance(item, str):
            raise KeyError('Annotation identifier must be an integer indexing Annotation groups or string representing a category name.')

        for idx, AnnotationTypeKey in enumerate(self.AnnotationTypesKeys):
            if AnnotationTypeKey == item:
                item = idx

        self.Annotations.Annotation[item].Element.clear()
        del self.Annotations.Annotation[item]

    def dump(self, path=None):
        def indent(elem, level=0): # creates new_line and indents in the xml file
            i = "\r\n" + level*"  "
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + "  "
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for elem in elem:
                    indent(elem, level+1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i

        AnnotationDataElement = deepcopy(self.AnnotationData.Element) # deep copy - indent not in the original file
        for idx, Element in enumerate(AnnotationDataElement):
            if Element.__len__() == 0:
                AnnotationDataElement[idx].clear()
                del AnnotationDataElement[idx]

        indent(AnnotationDataElement) # indents
        xml = ET.ElementTree(AnnotationDataElement) # Tree to write
        if isinstance(path, type(None)):
            path = self.AnnotationData._file_path

        xml.write(path, encoding='utf-8', xml_declaration=True) # dump file

def parse_CyberPSG_Annotations_xml(path):
    Annotations = CyberPSGFile(path)
    annotationTypes = Annotations.get_annotation_types()
    annotations = Annotations.get_annotations()
    dfAnnotations = pd.DataFrame(annotations)
    return dfAnnotations, annotationTypes














import torch
import numpy as np
import scipy.signal as signal
from best.signal_generation.DCGAN import models_DCGAN
from best.dbs import ArtifactGenerator

from mef_tools.io import MefReader
from best.signal import buffer, get_datarate

class RCSDataset:
    def __init__(self, path='/mnt/Helium/filip/Projects/2020_Sleep_Analysis/2022_sleep_architecture_only/M1/M1_nostim_1575446160.0/M1_nostim_1575446160_250Hz_filtered_e0-e3_e12-e13_e4-e5_e8-e11.mefd'):
        Rdr = MefReader(path)
        data = []
        for ch in Rdr.channels:
            if not 'cc' in ch:
                x = Rdr.get_data(ch)
                xb = buffer(x, 250, 3)
                dr = get_datarate(xb)
                xb = xb[dr > 0.95]
                xb[np.isnan(xb)] = np.nanmean(xb)
                xb500 = signal.resample(xb, xb.shape[1]*2, axis=1)
                data += [xb500]

        data = np.concatenate(data, 0)
        self.data = (data - data.mean(1).reshape(-1, 1)) / data.std(1).reshape(-1, 1)

        self._len = 12800000

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        idx = np.random.randint(self.data.shape[0])
        return self.data[idx]

class StimArtifactDataset:
    def __init__(self, sig_len=60, use_models=['MultiCenteriEEG_pathology', 'MultiCenteriEEG_physiology'], fs=500, use_artifacts=['RCS'], device='cpu'):
        self._device = device
        # print(device)
        # print(device)
        # print(device)
        print(device)
        self.SigGenerators = [models_DCGAN[k].to(device) for k in use_models]
        self.ArtGenerators = [ArtifactGenerator(k, 500) for k in use_artifacts]
        self._len = 1000000
        self.sig_len = sig_len
        self.fs = fs


    def to(self, device):
        self.SigGenerators = [sg.to(device) for sg in self.SigGenerators]

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        self.to(device)

    def __getitem__(self, item):
        idx_sig_gen = item % self.SigGenerators.__len__()
        idx_art_gen = item % self.ArtGenerators.__len__()

        X_orig = self.SigGenerators[idx_sig_gen].generate_signal(n_batch=1, n_seconds=self.sig_len, momentum=0.1).squeeze(0)
        # X_orig = X_orig - np.nanmean(X_orig)

        if item % 2 == 0:
            X_art, Y_art = self.ArtGenerators[idx_art_gen].get_signal_random(n_batch=1, n_length=X_orig.shape[-1])
        else:
            f = 10 ** (np.random.rand()*(1.65-0.3)+0.3) # 2 - 45 Hz
            a = 10 ** (np.random.rand()*(1--0.3)+-0.3) # 0.01 - 10
            X_art, Y_art = self.ArtGenerators[idx_art_gen].get_signal_periodic(n_batch=1, n_length=X_orig.shape[-1], freq=f, ampl_rel=a)

        # b, a = signal.butter(np.random.randint(1,3), np.random.rand()*80 + 100, fs=self.fs)
        # X_art = signal.filtfilt(b, a, X_art).copy()

        X_art = torch.tensor(X_art, dtype=torch.float32).squeeze(0)
        Y_art = torch.tensor(Y_art, dtype=torch.float32).squeeze(0)
        X_art = X_orig + X_art

        X_orig = X_orig - np.nanmean(X_art)
        X_art = X_art - np.nanmean(X_art)

        if self.fs == 250:
            return X_orig[::2], X_art[::2], Y_art[::]

        return X_orig, X_art, Y_art

    def __len__(self):
        return self._len

class StimArtifactDataset_RCS:
    def __init__(self, sig_len=60, use_models=[], fs=500, use_artifacts=['RCS'], device='cpu'):
        self._device = device
        # print(device)
        # print(device)
        # print(device)
        # print(device)
        # self.SigGenerators = [models_DCGAN[k].to(device) for k in use_models]
        self.SigDat = RCSDataset()
        self.ArtGenerators = [ArtifactGenerator(k, 500) for k in use_artifacts]
        self._len = 1000000
        self.sig_len = sig_len
        self.fs = fs


    def to(self, device):
        self.SigGenerators = [sg.to(device) for sg in self.SigGenerators]

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        self.to(device)

    def __getitem__(self, item):
        # idx_sig_gen = item % self.SigGenerators.__len__()
        idx_art_gen = item % self.ArtGenerators.__len__()

        X_orig = torch.tensor(self.SigDat[item].reshape(1, -1).copy(), device=self._device).float()
        #self.SigGenerators[idx_sig_gen].generate_signal(n_batch=1, n_seconds=self.sig_len, momentum=0.1).squeeze(0)
        X_orig = X_orig - np.nanmean(X_orig)


        if item % 2 == 0:
            X_art, Y_art = self.ArtGenerators[idx_art_gen].get_signal_random(n_batch=1, n_length=X_orig.shape[-1])
        else:
            f = 10 ** (np.random.rand()*(1.2-0.3)+0.3) # 2 - 45 Hz
            a = 10 ** (np.random.rand()*(1--0.3)+-0.3) # 0.01 - 10
            X_art, Y_art = self.ArtGenerators[idx_art_gen].get_signal_periodic(n_batch=1, n_length=X_orig.shape[-1], freq=f, ampl_rel=a)

        X_art = torch.tensor(X_art, dtype=torch.float32).squeeze(0)
        Y_art = torch.tensor(Y_art, dtype=torch.float32).squeeze(0)
        X_art = X_orig + X_art
        X_art = X_art - np.nanmean(X_art)
        if self.fs == 250:
            return X_orig[:, ::2], X_art[:, ::2], Y_art[:, ::2]
        #
        # b, a = signal.butter(np.random.randint(1,7), np.random.rand()*80 + 100, fs=self.fs)
        # X_orig = signal.filtfilt(b, a, X_orig)
        # X_art = signal.filtfilt(b, a, X_art)

        return X_orig.copy(), X_art.copy(), Y_art

    def __len__(self):
        return self._len







# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from scipy.stats import gaussian_kde
from best.modules import ZScoreModule
from sklearn.metrics import precision_recall_curve, auc, roc_auc_score, average_precision_score
from sklearn.neighbors import KernelDensity

def kl_divergence(mu1, std1, mu2, std2):
    """
    Parametric KL-Divergence between 2 normal 1-D distributions.

    `Normal Distribution <https://en.wikipedia.org/wiki/Normal_distribution>`_


    """
    return 0.5 * ((std1/std2)**2 + ((mu2-mu1)**2 / std2**2) -1 + 2*np.log(std2/std1))


def kl_divergence_mv(mu1, var1, mu2, var2):
    """
    Multidimensional parametric KL-Divergence between 2 normal distributions.

    `KL-Divergence <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Multivariate_normal_distributions>`_

    `Trace <https://en.wikipedia.org/wiki/Trace_(linear_algebra)>`_

    """
    return 0.5 * ((np.trace(np.dot(np.linalg.inv(var2), var1))) + np.dot(np.dot((mu2 - mu1), np.linalg.inv(var2)), (mu2-mu1).T) - mu1.shape[1] + np.log(np.linalg.det(var2)/np.linalg.det(var1)))[0, 0]


def combine_gauss_distributions(mu1, std1, N1, mu2, std2, N2):
    """
    Recalculates a normal 1-D distribution given two subsets of data.
    """

    c1 = N1 / (N1 + N2)
    c2 = N2 / (N1 + N2)
    mu_combined = (mu1 * c1) + (mu2 * c2)
    std_combined = np.sqrt(
        (N1*std1**2 + N2*std2**2 + N1*((mu1 - mu_combined)**2) + N2*((mu2 - mu_combined)**2)) / (N1+N2)
    ) #
    # np.sqrt((N1*(std1**2) + N2*(std2**2) + N1*N2*(mu2-mu1)**2/(N1+N2)) / (N1+N2)) # https://prod-ng.sandia.gov/techlib-noauth/access-control.cgi/2008/086212.pdf - 1.4
    return mu_combined, std_combined


def combine_mvgauss_distributions(mu1, var1, N1, mu2, var2, N2):
    """
    Recalculates a normal n-D distribution given two subsets of data.
    """
    c1 = N1 / (N1 + N2)
    c2 = N2 / (N1 + N2)
    mu_combined = (mu1 * c1) + (mu2 * c2)
    var_combined = (N1*(var1) + N2*(var2) + N1*N2*(mu2-mu1)**2/(N1+N2)) / (N1+N2) # np.sqrt((N1*(std1**2) + N2*(std2**2) + N1*N2*(mu2-mu1)**2/(N1+N2)) / (N1+N2)) # https://prod-ng.sandia.gov/techlib-noauth/access-control.cgi/2008/086212.pdf
    for k1 in range(var_combined.shape[0]):
        for k2 in range(k1+1, var_combined.shape[0]):
            var_combined[k2, k1] = var_combined[k1, k2]
    return mu_combined, var_combined


def kl_divergence_nonparametric(pk, qk):
    """
    Calculates non-parametric KL-Divergence between two 1-D distributions given by 2 histograms with same bins.
    """
    l_ = pk / qk
    barr = (~np.isinf(l_)) & (~np.isnan(l_))
    return np.nansum(pk[barr] * np.log(l_[barr]))


def get_class_count(Y, classes=None):
    """
    Returns a number of class appearance in the labels
    """
    if isinstance(classes, type(None)):
        classes = np.unique(Y)
        classes.sort()
    return dict([(cl, Y==cl) for cl in classes])


def compare_datasets(dataset, states=['AWAKE', 'N2', 'N3', 'REM']):
    """
    Compares dataset consistency using KDE and prec-recall curves

    Parameters
    ----------
    dataset: dict
        dict where the key is name of dataset for comparison; each dataset is represented by dict with 2 variables X, Y

    Returns
    -------
    dataset: dict
    """
    # AUPRC, AUROC, likelihood
    # sturcture
    # dataset_name - measure_name: AUPRC, AUROC, av likelihood_per_class, av_likelihood, class_frequency - class

    scores = {}
    for idx1, k1 in enumerate(dataset.keys()):
        for idx2, k2 in enumerate(dataset.keys()):
            x1 = dataset[k1]['X']
            y1 = dataset[k1]['Y']
            x2 = dataset[k2]['X']
            y2 = dataset[k2]['Y']

            #nm = k1 + ' - ' + k2

            ZS = ZScoreModule(trainable=True)
            x1 = ZS.fit_transform(x1)
            x2 = ZS.transform(x2)

            kde = KernelDensity().fit(x1)
            #y1_ = kde.score_samples(x1)
            y2_ = kde.score_samples(x2)

            if not k1 in scores.keys(): scores[k1] = {}

            if not k2 in scores[k1].keys():
                scores[k1][k2] = {}
                scores[k1][k2]['loglikelihood'] = {}
                scores[k1][k2]['auroc'] = {}
                #scores[k1][k2]['auprc'] = {}
                scores[k1][k2]['ap'] = {}

            scores[k1][k2]['loglikelihood']['all'] = y2_.mean()

            for cl in states:
                kde = KernelDensity().fit(x1[y1 == cl])
                sc = kde.score_samples(x2)
                #pr, rc, thresholds = precision_recall_curve(y2 == cl, sc)
                #pr = pr[1:]
                #rc = rc[1:]
                #auprc = auc(rc, pr)
                auroc = roc_auc_score(y2 == cl, sc)
                ap = average_precision_score(y2 == cl, sc)
                #scores[k1][k2]['auprc'][cl] = auprc
                scores[k1][k2]['auroc'][cl] = auroc
                scores[k1][k2]['ap'][cl] = ap
                scores[k1][k2]['loglikelihood'][cl] = sc.mean()
    return scores


# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import numpy as np
import pandas as pd
import datetime
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from dateutil import tz
from copy import deepcopy
from tqdm import tqdm

from best.hypnogram.visualisation import plot_hypnogram
from best.hypnogram.utils import merge_annotations, filter_by_key


"""
Tools for analyzing hypnograms such as number of cycles, sleep time etc.
"""


def get_hypnogram_datarate(df):
    #return (df['start'][1:].reset_index(drop=True) - df['end'][:-1].reset_index(drop=True)).sum().seconds
    return ((df['duration'].sum()) / (df.iloc[-1]['end'] - df.iloc[0]['start']).seconds)


def get_fell_asleep_time(df, t_sleep_check=60, t_awake_threshold=10, awake_tag='AWAKE', sleep_cycle_tags=['REM', 'N1', 'N2', 'N3']):
    df = filter_by_key(df, 'annotation', 'Arrousal')
    # parameters
    t_sleep = datetime.timedelta(minutes=t_sleep_check) # interval since 1st asleep checked
    t_awake = datetime.timedelta(minutes=t_awake_threshold) # length of all awake cycles during the interval defined by t_sleep since the beginning of the sleep

    # get first asleep
    awake_to_sleep_changes = np.where(
        np.array([df.annotation[0] in sleep_cycle_tags] +
                 [(df.annotation[k-1] in awake_tag) and (df.annotation[k] in sleep_cycle_tags) for k in range(1, df.__len__())
                  ]))[0]

    fell_asleep_time = df.iloc[0].start
    for idx in awake_to_sleep_changes:
        start = df.start[idx]
        all_starts = df.start
        t_sleep_window_df = df.loc[(start <= all_starts) & (all_starts < (start + t_sleep))]
        time_awake = t_sleep_window_df.duration[t_sleep_window_df.annotation == awake_tag].sum()
        if time_awake < t_awake.seconds:
            fell_asleep_time = start
            break
    return fell_asleep_time


def get_awakening_time(df, t_awake_threshold=90, t_sleep_threshold=10, awake_tag='AWAKE', sleep_cycle_tags=['REM', 'N2', 'N3']):
    df = filter_by_key(df, 'annotation', 'Arrousal')
    t_awake = datetime.timedelta(minutes=t_awake_threshold)
    t_sleep = datetime.timedelta(minutes=t_sleep_threshold)

    sleep_to_awake_changes = np.where(
        np.array([df.annotation[0] in sleep_cycle_tags] +
                 [(df.annotation[k-1] in sleep_cycle_tags) and (df.annotation[k] in awake_tag) for k in range(1, df.__len__())]
                 ))[0]

    if df.iloc[-1].annotation == awake_tag:
        last_awake_time = df.iloc[sleep_to_awake_changes[-1]].start
    else:
        last_awake_time = df.iloc[-1].end

    for idx in sleep_to_awake_changes:
        start = df.start[idx]
        all_starts = df.start
        t_awake_window_df = df.loc[(start <= all_starts) & (all_starts < (start + t_awake))]
        time_awake = t_awake_window_df.duration[t_awake_window_df.annotation == awake_tag].sum()
        time_asleep = get_time_by_key(t_awake_window_df, ['N1', 'N2', 'N3', 'REM'])
        if time_awake >= t_awake.seconds and time_asleep <= t_sleep_threshold:
            last_awake_time = start
            break
    return last_awake_time


def is_sleep_complete(df, awake_tag='AWAKE'):
    return df.iloc[0].annotation == awake_tag == df.iloc[-1].annotation


def get_rem_latency(df, rem_tag='REM', awake_tag='AWAKE'):
    first_rem_start = df.loc[df.annotation == rem_tag].reset_index(drop=True).iloc[0].start
    fall_asleep_start = get_fell_asleep_time(df)
    df_rem = df.loc[(df.end <= first_rem_start) & (df.annotation == awake_tag)].reset_index(drop=True)
    if df_rem.__len__() == 0:
        last_awake_end = fall_asleep_start
    else:
        last_awake_end = df_rem.iloc[-1].end

    return {'last_awake': first_rem_start - last_awake_end, 'fall_asleep': first_rem_start - fall_asleep_start}


def get_number_of_sleep_stages(df, tags ='REM', delay=30):
    if isinstance(tags, str):
        tags = [tags]

    delay = datetime.timedelta(minutes=delay)
    bool_idxes = np.ones(df.__len__(), dtype=bool)
    for tag in tags:
        bool_idxes = (bool_idxes) & (df.annotation == tag)

    df = df.loc[bool_idxes].reset_index(drop=True)

    stage_df = pd.DataFrame()
    for idx, row in enumerate(df.iterrows()):
        if idx == 0:
            stage_df = stage_df.append(row[1], ignore_index=True)
        else:
            if (row[1].start - stage_df.iloc[-1].end).seconds >= delay.seconds:
                stage_df = stage_df.append(row[1], ignore_index=True)

    return (stage_df.annotation == tag).sum()




def get_number_of_awakenings(df, awake_tag='AWAKE', n1_tag='N1', sleep_tags=['N2', 'N3', 'REM']):
    awake_bool = df.annotation == awake_tag
    n1_bool = df.annotation == n1_tag
    sleep_bool = np.zeros_like(n1_bool, dtype=bool)
    for tag in sleep_tags:
        sleep_bool = (sleep_bool) | (df.annotation == tag)

    awake_n1_bool = (awake_bool) | (n1_bool)

    n_awakenings = 0
    sleep_happened = False
    awake_happened = False
    for k in range(sleep_bool.shape[0]):
        if sleep_bool[k] == True:
            sleep_happened = True

        if awake_bool[k] == True:
            awake_happened = True
        else:
            if not awake_n1_bool[k] == True:
                awake_happened = False

        if sleep_happened == True and awake_happened == True:
            n_awakenings += 1
            sleep_happened = False
            awake_happened = False
    return n_awakenings


def get_time_by_key(df, key):
    if isinstance(key, list):
        value = 0
        for single_key in key:
            value += (df.duration[(df.annotation == single_key)]).sum()
        return value
    else:
        return (df.duration[(df.annotation == key)]).sum()


def get_stage_times(df, keys):
    return [get_time_by_key(df, key) for key in keys]


def get_stage_times_dataset(hypnograms:list, keys, verbose=True):
    if verbose:
        return pd.DataFrame([dict([(state, get_time_by_key(hyp, state)) for state in keys]) for hyp in tqdm(hypnograms)])
    return pd.DataFrame([dict([(state, get_time_by_key(hyp, state)) for state in keys]) for hyp in hypnograms])


def score_night(df, plot=False):
    df = filter_by_key(df, 'annotation', 'Arrousal')
    df = merge_annotations(df)

    fell_asleep_time = get_fell_asleep_time(df)
    awakening_time = get_awakening_time(df)
    sleep_complete = is_sleep_complete(df)

    sleep_df = df.loc[(df.start >= fell_asleep_time) & (df.start < awakening_time)].reset_index(drop=True)

    n_complete_sleep_cycles = get_number_of_sleep_stages(sleep_df, tags='REM', delay=30)
    n_awakenings = get_number_of_awakenings(sleep_df)
    rem_latency = get_rem_latency(df)


    n1_sleep_time = get_time_by_key(sleep_df, 'N1')
    n2_sleep_time = get_time_by_key(sleep_df, 'N2')
    n3_sleep_time = get_time_by_key(sleep_df, 'N3')
    rem_sleep_time = get_time_by_key(sleep_df, 'REM')
    awake_sleep_time = get_time_by_key(sleep_df, 'AWAKE')


    if plot == True:
        plot_hypnogram(df)
        plt.stem([fell_asleep_time, awakening_time], [7, 7], linefmt='r', markerfmt='or', basefmt='r')

    return {
        'sleep_complete': sleep_complete,
        'fell_asleep_time': fell_asleep_time,
        'rem_latency_fell_asleep': rem_latency['fall_asleep'].seconds,
        'rem_latency_last_awake': rem_latency['last_awake'].seconds,
        'awakening_time': awakening_time,
        'n_complete_sleep_cycles': n_complete_sleep_cycles,
        'n_awakenings': n_awakenings,
        'n1_sleep_time': n1_sleep_time,
        'n2_sleep_time': n2_sleep_time,
        'n3_sleep_time': n3_sleep_time,
        'rem_sleep_time': rem_sleep_time,
        'awake_sleep_time': awake_sleep_time
    }


def print_sleep_score(score):
    total_sleep_time =(score['awakening_time'] - score['fell_asleep_time']).seconds
    hours, remainder = divmod(total_sleep_time, 3600)
    minutes, seconds = divmod(remainder, 60)

    rem_lat1 = score['rem_latency_last_awake'].seconds
    hrs_rem1, remainder = divmod(rem_lat1, 3600)
    mins_rem1, secs_rem1 = divmod(remainder, 60)

    rem_lat2 = score['rem_latency_fell_asleep'].seconds
    hrs_rem2, remainder = divmod(rem_lat2, 3600)
    mins_rem2, secs_rem2 = divmod(remainder, 60)

    non_REM = score['n3_sleep_time'] + score['n2_sleep_time'] + score['n1_sleep_time']

    print('Sleep Complete: ', score['sleep_complete'])
    print('Falling asleep: ', score['fell_asleep_time'].strftime('%H:%M:%S'))
    print('Awakening: ', score['awakening_time'].strftime('%H:%M:%S'))
    print('Total Sleep Time: {:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds)))
    print('Rem latency - last_awake: {:02}:{:02}:{:02}'.format(int(hrs_rem1), int(mins_rem1), int(secs_rem1)))
    print('Rem latency - fall_asleep: {:02}:{:02}:{:02}'.format(int(hrs_rem2), int(mins_rem2), int(secs_rem2)))

    print('Number of sleep cycles: ', score['n_complete_sleep_cycles'])
    print('Number of awakenings', score['n_awakenings'])
    print()
    print('Sleep-time non-REM')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(non_REM),  non_REM/total_sleep_time))
    print('Sleep-time REM')
    print()
    print('Sleep-time REM')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(score['rem_sleep_time']),  score['rem_sleep_time']/total_sleep_time))
    print()
    print('Sleep-time awake')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(score['awake_sleep_time']),  score['awake_sleep_time']/total_sleep_time))
    print()
    print('Sleep-time N1')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(score['n1_sleep_time']),  score['n1_sleep_time']/total_sleep_time))
    print()
    print('Sleep-time N2')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(score['n2_sleep_time']),  score['n2_sleep_time']/total_sleep_time))
    print()
    print('Sleep-time N3')
    print('Absolute: {0}  Relative: {1:0.3f}'.format(int(score['n3_sleep_time']),  score['n3_sleep_time']/total_sleep_time))


def get_transition_counts(hyp, states=['AWAKE', 'N1', 'N2', 'N3', 'REM']):
    states = np.array(states)
    matrix = np.zeros((states.__len__(), states.__len__()))
    for k in range(hyp.__len__()-1):
        s1 =  hyp['annotation'][k]
        s2 =  hyp['annotation'][k+1]
        if s1 in states and s2 in states:
            idx1 = np.where(states == hyp['annotation'][k])[0][0]
            idx2 = np.where(states == hyp['annotation'][k+1])[0][0]
            matrix[idx1, idx2] += 1
    return matrix


def get_transition_matrix(hyp, states=['AWAKE', 'N1', 'N2', 'N3', 'REM']):
    m = get_transition_counts(hyp, states)
    m = m / m.sum(axis=1).reshape(-1, 1)
    m[np.isnan(m)] = 0
    return m


def get_transition_matrix_dataset(hypnograms, states=['AWAKE', 'N1', 'N2', 'N3', 'REM']):
    ms = []
    for hyp in hypnograms:
        m = get_transition_counts(hyp, states)
        m = m / m.sum(axis=1).reshape(-1, 1)
        ms += [m]
    ms = np.array(ms)
    return np.nanmean(ms, axis=0), np.nanstd(ms, axis=0)


def transition_matrix_to_change_matrix(m):
    m = deepcopy(m)
    np.fill_diagonal(m, 0)
    m = m / m.sum(axis=1).reshape(-1, 1)
    m[np.isnan(m)] = 0
    return m


def valid_dataset_index_by_duration(hypnograms:list, filt_dict:dict):
    valid_hypnograms = []
    for idx, hyp in enumerate(hypnograms):
        if sum([hyp['duration'].sum() >= v for k, v in filt_dict.items() if k in hyp['annotation'].unique()]) == filt_dict.keys().__len__():
            valid_hypnograms += [idx]
    return valid_hypnograms









# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import datetime
from io import StringIO
import xml.etree.ElementTree as ET
import os
import pandas as pd

from io import StringIO
import xml.etree.ElementTree as ET
import os
import pandas as pd
from tqdm import tqdm

from mef_tools.io import MefReader

from best.hypnogram.CyberPSG import CyberPSGFile, CyberPSG_XML_Writter
from best.hypnogram.NSRR import NSRRSleepFile
from best.hypnogram.utils import time_to_utc, create_duration, tile_annotations
from copy import deepcopy



"""
Tools for loading and saving of hypnogram files.
"""

_hypnogram_colors = {
    'AWAKE': '#e7b233',
    'WAKE': '#e7b233',
    'Arousal': '#d44b05',
    'REM': '#3500d3',
    'N1': '#2bc7c4',  # 2b7cc7
    'N2': '#2b5dc7',
    'N3': '#000000',
    'SLP': '#2b5dc7',
    'UNKNOWN': '#eaeded',
    'C_N2Huijie_1154_19_07_04': '#FF548ED5',
    'C_AWAKEHuijie_1154_19_07_04': '#FFFFFF00',
    'C_REMHuijie_1154_19_07_04': '#FF7030A0',
    'AWAKE_b': '#e7b233',
    'REM_b': '#3500d3',
    'N1_b': '#2bc7c4',  # 2b7cc7
    'N2_b': '#2b5dc7',
    'N3_b': '#000000',
    'AWAKE_dr': '#e7b233',
    'REM_dr': '#3500d3',
    'N1_dr': '#2bc7c4',  # 2b7cc7
    'N2_dr': '#2b5dc7',
    'N3_dr': '#000000',
    'AWAKE_aisc': '#e7b233',
    'REM_aisc': '#3500d3',
    'N1_aisc': '#2bc7c4',  # 2b7cc7
    'N2_aisc': '#2b5dc7',
    'N3_aisc': '#000000',
    'N_aisc': '#000000',
    'SLP_aisc': '#000000',
    'AWAKE_PiesPro': '#e7b233',
    'REM_PiesPro': '#3500d3',
    'N1_PiesPro': '#2bc7c4',  # 2b7cc7
    'N2_PiesPro': '#2b5dc7',
    'N3_PiesPro': '#000000',
    'N_PiesPro': '#000000',
    'SLP_PiesPro': '#000000',
    'seizure_PiesPro': '#000000',

    'AWAKE_best': '#e7b233',
    'REM_best': '#3500d3',
    'N1_best': '#2bc7c4',  # 2b7cc7
    'N2_best': '#2b5dc7',
    'N3_best': '#000000',
    'N_best': '#000000',
    'SLP_best': '#000000',
    'seizure_best': '#000000',

    'left': '#3500d3',
    'right': '#2bc7c4',  # 2b7cc7
    'both': '#2b5dc7',
    'self_report': '#000000',
    'cluster': '#000000',

    'IED': '#000000',
    'IED_best': '#000000',

    'left_PiesPro': '#3500d3',
    'right_PiesPro': '#2bc7c4',  # 2b7cc7
    'both_PiesPro': '#2b5dc7',
    'self_report_PiesPro': '#000000',
    'cluster_PiesPro': '#000000',

    'left_best': '#3500d3',
    'right_best': '#2bc7c4',  # 2b7cc7
    'both_best': '#2b5dc7',
    'self_report_best': '#000000',
    'cluster_best': '#000000',

    'Sleep stage W': '#e7b233',
    'Sleep stage R': '#3500d3',
    'Sleep stage N1': '#2bc7c4',  # 2b7cc7
    'Sleep stage N2': '#2b5dc7',
    'Sleep stage N3': '#000000',

    'Sleep stage W_best': '#e7b233',
    'Sleep stage R_best': '#3500d3',
    'Sleep stage N1_best': '#2bc7c4',  # 2b7cc7
    'Sleep stage N2_best': '#2b5dc7',
    'Sleep stage N3_best': '#000000',
}


def load_CyberPSG(path, tile=None, verbose=True):
    if isinstance(path, list):
        return _load_CyberPSG_dataset(path, tile, verbose)
    else:
        return _load_CyberPSG(path, tile)


def _load_CyberPSG(path, tile=None):
    if not os.path.isfile(path):
        raise FileNotFoundError('[FILE ERROR]: File not found ' + path)
    fid = CyberPSGFile(path)
    annotations = fid.get_hypnogram()
    df = pd.DataFrame(annotations)
    df = create_duration(df)
    if not isinstance(tile, type(None)):
        if (df.duration > tile).sum() > 0:
            df = tile_annotations(df, tile)

    for k in df.annotation.unique():
        if k[-5:] == '_aisc':
            df.loc[df.annotation == k, 'annotation'] = k[:-5]

        if k[-8:] == '_PiesPro':
            df.loc[df.annotation == k, 'annotation'] = k[:-8]
    return df


def _load_CyberPSG_dataset(paths: list, tile=None, verbose=True):
    if verbose:
        print('Loading hypnogram Dataset')
        return [_load_CyberPSG(pth, tile) for pth in tqdm(paths)]
    else:
        return [_load_CyberPSG(pth, tile) for pth in paths]


def save_CyberPSG(path, df):
    #TODO: Do Tests
    #TODO: Implement annotation groups etc

    fid = CyberPSG_XML_Writter(path)
    annotation_group = 'Import_best'
    annotation_types = list(df['annotation'].unique())
    df = deepcopy(df)

    if 'channel' in df.keys():
        df.loc[df.channel.isnull(), 'channel'] = ''

    df = time_to_utc(df)

    fid.add_AnnotationGroup(annotation_group, uuid_='00000000-0000-0000-0000-000000000001')
    for atype in annotation_types:
        if atype in _hypnogram_colors.keys():
            fid.add_AnnotationType(atype, groupAssociationId=annotation_group, color=_hypnogram_colors[atype])
        else:
            fid.add_AnnotationType(atype, groupAssociationId=annotation_group)

    for idx, row in df.iterrows():
        write_channel = False
        if 'channel' in row.keys():
            if row['channel']:
                write_channel = True

        if write_channel:
            fid.add_Annotation(row['start'], row['end'], AnnotationTypeId=row['annotation'], channelName=row['channel'], type='ChannelAnnotation')
        else:
            fid.add_Annotation(row['start'], row['end'], AnnotationTypeId=row['annotation'], type='GlobalAnnotation')

    fid.dump()


def load_NSRR(path):
    return NSRRSleepFile(path).get_hypnogram()


def load_MefFile(mefid: MefReader):
    hyp = pd.DataFrame(mefid.get_annotations())
    hyp['duration'] = hyp['duration']/1e6
    hyp = hyp.drop(columns='type')

    hyp['start'] = hyp['time']/1e6
    hyp = hyp.drop(['time'], axis=1)

    hyp['end'] = hyp['start'] + hyp['duration']

    hyp['annotation'] = hyp['text']
    hyp = hyp.drop(['text'], axis=1)

    hyp = hyp[['annotation', 'start', 'end', 'duration']]

    return hyp

import numpy as np







# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os
import numpy as np
from shutil import rmtree
from time import sleep


def create_folder(PATH_TO_CREATE):
    """
    Creates a folder on PATH_TO_CREATE position.
    """
    from os import mkdir
    mkdir(PATH_TO_CREATE)

def remove_folder(PATH_TO_REMOVE):
    """
    Creates a folder on PATH_TO_CREATE position.
    """
    try: rmtree(PATH_TO_REMOVE)
    except: pass
    sleep(0.05)

def get_files(path, endings=None, creation_time=False):
    """File list generator.

        For each file in path directory and subdirectories with endings from endings_tuple returns path in a list variable

        path is a string, the path to the directory where you wanna search specific files.

        endings_tuple is a tuple of any length

        The function returns paths to all files in folder and files in all subfolders as well.

        .. code-block:: python

            from best.files import get_files
            import pandas as pd

            path = "root/data"
            files_list = get_files(path, ('.jpg', '.jpeg'))
            files_pd = pd.DataFrame(files_list)

            def get_FID(x):
                temp = x['path'].split('\\')
                return temp[len(temp) - 1][0:-4]
            files_pd = files_pd.rename(columns={0: "path"})
            files_pd['fid'] = files_pd.apply(lambda x: get_FID(x), axis=1)
        """

    if isinstance(endings, str):
        endings = (endings, )

    if isinstance(endings, type(None)):
        data = [os.path.join(root, name)
                for root, dirs, files in os.walk(path)
                for name in files]

        data += [os.path.join(root, dir)
                for root, dirs, files in os.walk(path)
                for dir in dirs]
    else:
        data = [os.path.join(root, name)
                for root, dirs, files in os.walk(path)
                for name in files
                if name.endswith(endings)]

        data += [os.path.join(root, dir)
                for root, dirs, files in os.walk(path)
                for dir in dirs
                if dir.endswith(endings)]


    data = [file for file in data if not '._' in file]
    data.sort()

    if creation_time == False:
        return data

    creation_times = [os.stat(path).st_ctime for path in data]
    data = np.array(data)
    creation_times = np.array(creation_times)
    idx = creation_times.argsort()
    sorted_data = data[idx]
    sorted_creation_times = creation_times[idx]
    return sorted_data, sorted_creation_times

def get_folders(path):
    """
    Returns sub-folders present in the folder specified by the path.
    """
    return [name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name)) and ((not '.mefd' in name) and (not '.segd' in name) and (not '.timd' in name))]

def split_files(file_paths, ratio=0.5):
    """
    Randomly shuffles list of files and splits given by the ratio.
    """
    mix_idxes = np.random.permutation(file_paths.__len__())
    paths = list(np.array(file_paths)[mix_idxes])
    split_idx = int(np.round(ratio*paths.__len__()))
    return paths[:split_idx], paths[split_idx:]


# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import re
import os
import zmq
import pickle
import json
import numpy as np
import pandas as pd
import sqlalchemy as sqla
from tqdm import tqdm
from copy import deepcopy
from datetime import datetime
from sqlalchemy.pool import NullPool
from sshtunnel import SSHTunnelForwarder

from best.hypnogram.utils import create_day_indexes, time_to_timezone, time_to_timestamp, tile_annotations, create_duration
from best.cloud._db_connection_variables import *

import pandas as pd
from datetime import datetime



class DatabaseHandler:
    __version__ = '0.0.1'
    def __init__(self, sql_db_name, sql_host=None, sql_user=None, sql_pwd=None, sql_port=None, ssh_host=None, ssh_user=None, ssh_pwd=None, ssh_port=None):
        if isinstance(sql_host, type(None)):
            sql_host = IP_SQL
        if isinstance(sql_user, type(None)):
            sql_user = USER_SQL
        if isinstance(sql_pwd, type(None)):
            sql_pwd = PW_SQL
        if isinstance(sql_port, type(None)):
            sql_port = PORT_SQL

        if isinstance(ssh_host, type(None)):
            ssh_host = IP_SSH
        if isinstance(ssh_user, type(None)):
            ssh_user = USER_SSH
        if isinstance(ssh_pwd, type(None)):
            ssh_pwd = PW_SSH
        if isinstance(ssh_port, type(None)):
            ssh_port = PORT_SSH

        self._ssh_host = ssh_host
        self._ssh_user = ssh_user
        self._ssh_pwd = ssh_pwd
        self._ssh_port = ssh_port

        self._sql_host = sql_host
        self._sql_port = sql_port
        self._sql_user = sql_user
        self._sql_pwd = sql_pwd
        self._sql_db_name = sql_db_name

        self._sql_connection = None
        self._ssh_tunnel = None
        self._engine = None


        self.open()
        self._init_sql_engine()

    def _init_sql_engine(self):
        if self.check_ssh_connection():
            self._engine = sqla.create_engine(
                'mysql+pymysql://{}:{}@{}:{}/{}'.format(self._sql_user, self._sql_pwd, 'localhost', self._ssh_tunnel.local_bind_port, self._sql_db_name), poolclass=NullPool)
        else:
            self._engine = sqla.create_engine(
                'mysql+pymysql://{}:{}@{}:{}/{}'.format(self._sql_user, self._sql_pwd, self._sql_host, self._sql_port, self._sql_db_name), poolclass=NullPool)

    def _open_sql(self):
        self._sql_connection = self._engine.connect()

    def _close_sql(self):
        self._sql_connection.close()

    def check_sql_connection(self):
        self._open_sql()
        self._close_sql()
        return True

    def _open_ssh(self):
        self._ssh_tunnel = SSHTunnelForwarder(
            (self._ssh_host, int(self._ssh_port)),
            ssh_username=self._ssh_user,
            ssh_password=self._ssh_pwd,
            remote_bind_address=(self._sql_host, int(self._sql_port)))
        self._ssh_tunnel.start()

    def check_ssh_connection(self):
        if self._ssh_tunnel:
            return self._ssh_tunnel.is_active
        return False

    def _close_ssh(self):
        if self.check_ssh_connection():
            self._ssh_tunnel.close()

    def open(self):
        if self._ssh_host:
            self._open_ssh()
            self.check_ssh_connection()

        self._init_sql_engine()
        self.check_sql_connection()

    def close(self):
        self._close_sql()
        self._close_ssh()

    def check_connection(self):
        self.check_ssh_connection()
        return self.check_sql_connection()

    def __del__(self):
        self.close()

    @property
    def db_name(self):
        return self._sql_db_name

    @db_name.setter
    def db_name(self, name):
        self._sql_db_name = name
        self.check_connection()

class SessionFinder(DatabaseHandler):
    #TODO: Enable searching for signals between multiple session
    __version__ = '0.0.1'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def find_mef_session(self, patient_id, uutc_start, uutc_stop):
        if not isinstance(uutc_start, (int, float, np.int, np.float)):
            raise TypeError('uutc_start has to be of a number type - int or float. Data type ' + type(uutc_start) + ' found instead.')
        if not isinstance(uutc_stop, (int, float, np.int, np.float)):
            raise TypeError('uutc_stop has to be of a number type - int or float. Data type ' + type(uutc_stop) + ' found instead.')

        uutc_start = int(round(uutc_start*1e6))
        uutc_stop = int(round(uutc_stop*1e6))


        self._sql_connection = self._engine.connect()
        query = f"SELECT uutc_start, uutc_stop, session, fsamp, channels FROM {self._sql_db_name}.Sessions where id='{patient_id}' and uutc_start<='{uutc_start}' and uutc_stop>='{uutc_stop}' order by uutc_start desc"
        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        if df_data.__len__() > 0:
            return df_data.loc[0, 'session']

    def find_mef_session_bulk(self, patient_id, df):
        sessions = []
        self._sql_connection = self._engine.connect()

        for row in tqdm(list(df.iterrows())):
            row = row[1]
            if not isinstance(row['start'], (int, float, np.int, np.float)):
                raise TypeError('start column has to be of a number type - int or float. Data type ' + type(row['start']) + ' found instead.')
            if not isinstance(row['end'], (int, float, np.int, np.float)):
                raise TypeError('end column has to be of a number type - int or float. Data type ' + type(row['end']) + ' found instead.')

            uutc_start = int(round(row['start']*1e6))
            uutc_stop = int(round(row['end']*1e6))
            query = f"SELECT uutc_start, uutc_stop, session, fsamp, channels FROM {self._sql_db_name}.Sessions where id='{patient_id}' and uutc_start<='{uutc_start}' and uutc_stop>='{uutc_stop}' order by uutc_start desc"
            df_data = pd.read_sql(query, self._sql_connection)
            if df_data.__len__() > 0:
                sessions += [df_data.loc[0, 'session']]
            else:
                sessions += ['']
        self._sql_connection.close()
        return sessions

    @property
    def patient_ids(self):
        self._open_sql()
        query = f"SELECT DISTINCT id FROM {self._sql_db_name}.Sessions"
        unique_ids = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        return unique_ids['id'].to_list()

    def data_range(self, patient_id):
        self._open_sql()
        query = f"SELECT MIN(uutc_start), MAX(uutc_stop)  FROM {self._sql_db_name}.Sessions where id='{patient_id}'"
        unique_ids = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        return unique_ids.values[0]



class SleepClassificationModelDBHandler(DatabaseHandler):
    keys_pickle = ['classifier']
    keys_unalter = ['pid', 'kappa_score', 'name', 'class_name', 'package', 'package_version']

    """Can read and save models from and into DB"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def save_model(self, save_classifier):
        self._sql_connection = self._engine.connect()
        vals = {}
        for k, it in save_classifier.items():
            if not k in self.keys_unalter:
                if k in self.keys_pickle:
                    it = pickle.dumps(it)
                else:
                    it = json.dumps(it)

            vals[k] = it

        df = pd.DataFrame([vals])
        df.to_sql('sleep_classifier', con=self._sql_connection, if_exists='append', index=False)
        self._sql_connection.close()

    def load_classifier(self, classifier_id):
        self._sql_connection = self._engine.connect()
        query = f"SELECT * FROM UH3.sleep_classifier where id={classifier_id}"
        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        if df_data.__len__() > 0:
            data = {}
            for k, it in df_data.iloc[0].to_dict().items():
                if k != 'id':
                    if not k in self.keys_unalter:
                        if k in self.keys_pickle:
                            it = pickle.loads(it)
                        else:
                            it = json.loads(it)

                    data[k] = it
            return data


    def find_classifiers(self, pid, channels, name):
        self._sql_connection = self._engine.connect()
        query = f"SELECT id, channels, results, name, ampl_values, freq_values, remove_artifacts FROM UH3.sleep_classifier where pid='{pid}' and name='{name}'"
        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        data = []
        for idx, row in df_data.iterrows():
            drow = {}

            drow['id'] = row['id']
            drow['channels'] = json.loads(row['channels'])
            drow['name'] = row['name']
            drow['ampl_values'] = json.loads(row['ampl_values'])
            drow['freq_values'] = json.loads(row['freq_values'])
            drow['remove_artifacts'] = json.loads(row['remove_artifacts'])
            results = json.loads(row['results'])
            for metric, it in results.items():
                for state, value in it.items():
                    drow[f"{metric}_{state}"] = value

            if sum([ch in channels for ch in drow['channels']]) == drow['channels'].__len__():
                data += [drow]
        data = pd.DataFrame(data)
        return data

    @property
    def classifier_names(self):
        self._sql_connection = self._engine.connect()
        query = f"SELECT DISTINCT(name) FROM UH3.sleep_classifier"
        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()
        return [ch['name'] for idx, ch in df_data.iterrows()]

    # def load_model(self, patient_id, channel):
    #     self._sql_connection = self._engine.connect()
    #
    #     query = f"SELECT * FROM {self._sql_db_name}.sleep_classifier where patient_id='{patient_id}' and channel='{channel}'"
    #
    #     df_data = pd.read_sql(query, self._sql_connection)
    #     self._sql_connection.close()
    #     cls_string = df_data['classifier'][0]
    #     print(df_data.keys())
    #     return pickle.loads(cls_string), df_data.drop(columns='classifier')


class SleepDataDBHandler(DatabaseHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def write_data(self, df, patient_id, classifier_id):
        df = deepcopy(df)

        df['patient_id'] = patient_id
        df['classifier'] = classifier_id

        if 'annotation' in df.keys():
            df['sleep_stage'] = df['annotation']
            df = df.drop(['annotation'], axis=1)

        if 'start' in df.keys():
            df['start_uutc'] = df['start'] * 1e6
            df = df.drop(['start'], axis=1)

        if 'end' in df.keys():
            df['stop_uutc'] = df['end'] * 1e6
            df = df.drop(['end'], axis=1)


        self._sql_connection = self._engine.connect()
        df.to_sql('sleep_table', con=self._sql_connection, if_exists='append', index=False)
        self._sql_connection.close()

    def get_data(self, patient_id, start, end):
        start = int(round(start * 1e6))
        end = int(round(end * 1e6))

        self._sql_connection = self._engine.connect()
        query = f"SELECT sleep_stage, start_uutc, stop_uutc FROM {self._sql_db_name}.sleep_table where patient_id='{patient_id}' and start_uutc >= '{start}' and start_uutc <= '{end}'"
        df = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        df['end'] = df['stop_uutc'] / 1e6
        df = df.drop(['stop_uutc'], axis=1)

        df['start'] = df['start_uutc'] / 1e6
        df = df.drop(['start_uutc'], axis=1)
        df = create_duration(df)

        df['annotation'] = df['sleep_stage']
        df = df.drop(['sleep_stage'], axis=1)


        df = df[['annotation', 'start', 'end', 'duration']]
        return df


class SystemStateLoader(DatabaseHandler):

    def _get_sensing_channels(df, row_idx):
        det_channels = [df['sense_channel_name_' + str(idx)][row_idx] for idx in range(4)]
        det_channels = [chan_name.lower() for chan_name in det_channels if isinstance(chan_name, str)]
        return det_channels

    def _get_stim_channels(df, row_idx):
        stim_channels = []
        for idx in range(4):
            txt = df['current_group_prog' + str(idx) + 'electrodes'][row_idx]
            if isinstance(txt, str):
                stim_channels += re.findall(r'e\d{1,2}', txt.lower())
        return list(np.unique(stim_channels))

    def load_stimulation_info(self, patient_id, start_uutc):
        self._sql_connection = self._engine.connect()
        if np.log10(start_uutc) < 10: start_uutc = int(np.round(start_uutc*1e6))
        query = f"SELECT start_uutc, " \
                f"curr_Prog0AmpInMilliamps, curr_Prog1AmpInMilliamps, curr_Prog2AmpInMilliamps, curr_Prog3AmpInMilliamps, " \
                f"curr_program_pulsewidth0, curr_program_pulsewidth1, curr_program_pulsewidth2, curr_program_pulsewidth3, " \
                f"curr_rate_in_hz" \
                f" FROM {self.db_name}.System_Status_Processed WHERE patient_id='{patient_id}' and start_uutc<{start_uutc} ORDER BY start_uutc DESC LIMIT 1"

        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        if df_data.__len__() > 0:
            row = df_data.iloc[0]
            ampl = [row['curr_Prog'+str(k)+'AmpInMilliamps'] for k in range(4) if not isinstance(row['curr_Prog'+str(k)+'AmpInMilliamps'], type(None))]

            freq = row['curr_rate_in_hz']
            if isinstance(freq, type(None)): freq = 0

            if ampl.__len__() == 0: ampl = 0
            else: ampl = np.nanmax(ampl)

            pw = [row['curr_program_pulsewidth'+str(k)] for k in range(4) if not isinstance(row['curr_program_pulsewidth'+str(k)], type(None))]
            if pw.__len__() == 0: pw = 0
            else: pw = np.nanmax(pw)

            stim_info = {
                'freq': freq,
                'ampl': ampl,
                'pulsewidth': pw
            }
            return stim_info

    def load_stimulation_info_bulk(self, patient_id, df):
        df = time_to_timestamp(deepcopy(df))
        freq_ = []
        ampl_ = []
        pulsewidth_ = []

        self._sql_connection = self._engine.connect()


        for row in tqdm(range(df.__len__())):
            row = df.iloc[row]
            start_uutc = int(round(row['start'] * 1e6))
            query = f"SELECT start_uutc, " \
                    f"curr_Prog0AmpInMilliamps, curr_Prog1AmpInMilliamps, curr_Prog2AmpInMilliamps, curr_Prog3AmpInMilliamps, " \
                    f"curr_program_pulsewidth0, curr_program_pulsewidth1, curr_program_pulsewidth2, curr_program_pulsewidth3, " \
                    f"curr_rate_in_hz" \
                    f" FROM {self.db_name}.System_Status_Processed WHERE patient_id='{patient_id}' and start_uutc<{start_uutc} ORDER BY start_uutc DESC LIMIT 1"

            df_outp = pd.read_sql(query, self._sql_connection)

            if df_outp.__len__() > 0:
                row = df_outp.iloc[0]
                ampl = [row['curr_Prog'+str(k)+'AmpInMilliamps'] for k in range(4) if not isinstance(row['curr_Prog'+str(k)+'AmpInMilliamps'], type(None))]

                if ampl.__len__() == 0: ampl = 0
                else: ampl = np.nanmax(ampl)

                freq = row['curr_rate_in_hz']
                if isinstance(freq, type(None)): freq = 0


                pw = [row['curr_program_pulsewidth'+str(k)] for k in range(4) if not isinstance(row['curr_program_pulsewidth'+str(k)], type(None))]
                if pw.__len__() == 0: pw = 0
                else: pw = np.nanmax(pw)

            else:
                freq = ampl = pw = None


            freq_.append(freq)
            ampl_.append(ampl)
            pulsewidth_.append(pw)


        self._sql_connection.close()
        df['freq'] = freq_
        df['ampl'] = ampl_
        df['pulsewidth'] = pulsewidth_
        return df

    def load_stimulation_day_info(self, patient_id, start_uutc, stop_uutc):
        self._sql_connection = self._engine.connect()
        if np.log10(start_uutc) < 10: start_uutc = int(np.round(start_uutc*1e6))
        if np.log10(stop_uutc) < 10: stop_uutc = int(np.round(stop_uutc*1e6))

        query = f"SELECT start_uutc, " \
                f"curr_Prog0AmpInMilliamps, curr_Prog1AmpInMilliamps, curr_Prog2AmpInMilliamps, curr_Prog3AmpInMilliamps, " \
                f"curr_program_pulsewidth0, curr_program_pulsewidth1, curr_program_pulsewidth2, curr_program_pulsewidth3, " \
                f"curr_rate_in_hz" \
                f" FROM {self.db_name}.System_Status_Processed WHERE patient_id='{patient_id}' and start_uutc>{start_uutc} and start_uutc<{stop_uutc}"

        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        freqs = []
        ampls = []
        pws = []
        for row in df_data.iterrows():
            row = row[1]
            ampl = [row['curr_Prog'+str(k)+'AmpInMilliamps'] for k in range(4) if not isinstance(row['curr_Prog'+str(k)+'AmpInMilliamps'], type(None))]

            freq = row['curr_rate_in_hz']
            if isinstance(freq, type(None)): freq = 0

            if ampl.__len__() == 0: ampl = 0
            else: ampl = np.nanmax(ampl)

            pw = [row['curr_program_pulsewidth'+str(k)] for k in range(4) if not isinstance(row['curr_program_pulsewidth'+str(k)], type(None))]
            if pw.__len__() == 0: pw = 0
            else: pw = np.nanmax(pw)
            freqs += [freq]
            ampls += [ampl]
            pws += [pw]

        if freqs.__len__() == 0: freqs = None
        if ampls.__len__() == 0: ampls = None
        if pws.__len__() == 0: pws = None
        stim_info = {
            'freq': np.max(freqs),
            'ampl': np.max(ampls),
            'pulsewidth': np.max(pws)
        }
        return stim_info

        self._sql_connection = self._engine.connect()
        if np.log10(start_uutc) < 10: start_uutc = int(np.round(start_uutc * 1e6))
        if np.log10(stop_uutc) < 10: stop_uutc = int(np.round(stop_uutc * 1e6))

        query = f"SELECT start_uutc, " \
                f"curr_Prog0AmpInMilliamps, curr_Prog1AmpInMilliamps, curr_Prog2AmpInMilliamps, curr_Prog3AmpInMilliamps, " \
                f"curr_program_pulsewidth0, curr_program_pulsewidth1, curr_program_pulsewidth2, curr_program_pulsewidth3, " \
                f"curr_rate_in_hz" \
                f" FROM {self.db_name}.System_Status_Processed WHERE patient_id='{patient_id}' and start_uutc>{start_uutc} and start_uutc<{stop_uutc}"

        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        df_data['amplitude'] = df_data[['curr_Prog0AmpInMilliamps', 'curr_Prog1AmpInMilliamps', 'curr_Prog2AmpInMilliamps',
                                        'curr_Prog3AmpInMilliamps']].max(axis=1)
        df_data['frequency'] = df_data['curr_rate_in_hz']
        df_data['start'] = df_data['start_uutc'] / 1e6
        df_data['end'] = np.nan
        df_data.loc[:df_data.__len__() - 1, 'end'] = df_data.iloc[1:].reset_index(drop=True)['start']
        df_data = df_data.iloc[:-1]

        df_data = df_data[['frequency', 'amplitude', 'start', 'end']]
        df_data['duration'] = df_data['end'] - df_data['start']
        return df_data

    def load_stimulation_annotations(self, patient_id, start_uutc, stop_uutc):
        if np.log10(start_uutc) < 10: start_uutc = int(np.round(start_uutc * 1e6))
        if np.log10(stop_uutc) < 10: stop_uutc = int(np.round(stop_uutc * 1e6))

        query = f"SELECT start_uutc, " \
                f"curr_Prog0AmpInMilliamps, curr_Prog1AmpInMilliamps, curr_Prog2AmpInMilliamps, curr_Prog3AmpInMilliamps, " \
                f"curr_program_pulsewidth0, curr_program_pulsewidth1, curr_program_pulsewidth2, curr_program_pulsewidth3, " \
                f"curr_rate_in_hz, curr_stim_state, current_group, stim_mode" \
                f" FROM {self.db_name}.System_Status_Processed WHERE patient_id='{patient_id}' and start_uutc>{start_uutc} and start_uutc<{stop_uutc}"

        self._sql_connection = self._engine.connect()
        df_data = pd.read_sql(query, self._sql_connection)
        self._sql_connection.close()

        if df_data.__len__() > 0:
            df_data['amplitude'] = df_data[
                ['curr_Prog0AmpInMilliamps', 'curr_Prog1AmpInMilliamps', 'curr_Prog2AmpInMilliamps',
                 'curr_Prog3AmpInMilliamps']].max(axis=1)
            df_data['frequency'] = df_data['curr_rate_in_hz']
            df_data['start'] = df_data['start_uutc'] / 1e6
            df_data['end'] = np.nan
            df_data['pulse_width'] = df_data[['curr_program_pulsewidth0', 'curr_program_pulsewidth1', 'curr_program_pulsewidth2', 'curr_program_pulsewidth3']].max(axis=1)
            df_data.loc[:df_data.__len__() - 1, 'end'] = df_data.iloc[1:].reset_index(drop=True)['start']
            # df_data = df_data.iloc[:-1]
    
            df_data = df_data[['frequency', 'amplitude', 'start', 'end', 'pulse_width', 'curr_stim_state', 'current_group', 'stim_mode']]
            df_data['duration'] = df_data['end'] - df_data['start']
            
            return df_data
        
        else:
            return pd.DataFrame(columns=['frequency', 'amplitude', 'start', 'end', 'pulse_width', 'curr_stim_state', 'current_group', 'stim_mode'])
            

class ScientificDataLoader(DatabaseHandler):
    def get_impedance_min_max(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT MIN(start_uutc), MAX(stop_uutc)" \
                f" FROM {self._sql_db_name}.Impedance where patient_id='{patient_id}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and stop_uutc<={stop_timestamp}"
        self._open_sql()
        min_max = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        if not isinstance(min_max.iloc[0]['MIN(start_uutc)'], type(None)):
            min_max = [min_max.iloc[0]['MIN(start_uutc)'] / 1e6, min_max.iloc[0]['MAX(stop_uutc)'] / 1e6]
            return min_max
        else:
            return None

    def get_impedance_data(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"start_uutc, " \
                f"el_0, el_1, el_2, el_3, el_4, el_5, el_6, el_7, el_8, el_9, el_10, el_11, el_12, el_13, el_14, el_15 " \
                f"FROM {self._sql_db_name}.Impedance WHERE patient_id='{str(patient_id)}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        df['start_uutc'] /= 1e6
        return df


    def get_impedance_data_bipolar(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f" * FROM {self._sql_db_name}.Impedance_Bipol WHERE id='{str(patient_id)}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        df['start_uutc'] /= 1e6
        return df


    ############ Spike ##################
    def get_spike_rate_channels(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        self._open_sql()
        query = f"SELECT DISTINCT channel " \
                f"FROM {self._sql_db_name}.Automated_Events WHERE id='{str(patient_id)}' " \
                f"and ev_type='spike_rate' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        list_of_channels = np.array(df['channel'].unique())
        channels_by_location = {}
        for ch in list_of_channels:
            loc = self._get_channel_location(ch)
            if not loc in channels_by_location.keys():
                channels_by_location[loc] = []
            channels_by_location[loc].append(ch)
        return channels_by_location


    def get_spike_rate_data(self, patient_id=None, channel=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        if channel:
            query = f"SELECT " \
                    f"start_uutc, value  " \
                    f"FROM {self._sql_db_name}.Spike_Rate_Events WHERE id='{str(patient_id)}' " \
                    f"and channel='{channel}' " \
                    f"and start_uutc>={start_timestamp} " \
                    f"and start_uutc<={stop_timestamp}"
            self._open_sql()
            df = pd.read_sql(query, self._sql_connection)
            self._close_sql()
        else:
            query = f"SELECT " \
                    f"start_uutc, value, channel  " \
                    f"FROM {self._sql_db_name}.Spike_Rate_Events WHERE id='{str(patient_id)}' " \
                    f"and start_uutc>={start_timestamp} " \
                    f"and start_uutc<={stop_timestamp}"
            self._open_sql()
            df = pd.read_sql(query, self._sql_connection)
            self._close_sql()


        df['start_uutc'] /= 1e6
        df['start_uutc'] += 10*60
        return df

    def get_spike_n_detections(self,  patient_id=None, channel=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"COUNT(start_uutc)  " \
                f"FROM {self._sql_db_name}.Features WHERE id='{str(patient_id)}' " \
                f"and ev_type=\"spike\" " \
                f"and channel='{channel}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        return df['COUNT(start_uutc)'][0]


    def get_spike_detections(self,  patient_id=None, channel=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"start_uutc " \
                f"FROM {self._sql_db_name}.SpikesEvents WHERE id='{str(patient_id)}' " \
                f"and channel='{channel}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()

        df['start_uutc'] /= 1e6
        df['start_uutc'] += 10*60
        return df


    def get_seizures(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"start_uutc, stop_uutc " \
                f"FROM {self._sql_db_name}.Manual_Events WHERE id='{str(patient_id)}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp} " \
                f"and ev_type='seizure'  and source='eeg_review'"\
                #f"and source='eeg_review'"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        df['start_uutc'] /= 1e6
        # df['start_uutc'] += 10*60
        df['stop_uutc'] /= 1e6
        # df['stop_uutc'] += 10*60
        return df


    def get_seizure_probability(self, patient_id, seizures):
        seizures_ = []
        self._open_sql()
        for idx, szr in tqdm(list(seizures.iterrows())):
            query = f"SELECT " \
                    f"* " \
                    f"FROM {self._sql_db_name}.{patient_id}_prob WHERE " \
                    f"uutc_start>={szr.start_uutc * 1e6} " \
                    f"and uutc_stop<={szr.stop_uutc * 1e6} "
            # f"and source='eeg_review'"

            df = pd.read_sql(query, self._sql_connection)
            
            l = df.el1.to_numpy()
            r = df.el2.to_numpy()
            s_ = df.uutc_start.to_numpy()
            e_ = df.uutc_stop.to_numpy()

            fs = 1
            s_prob = szr.start_uutc
            e_prob = szr.stop_uutc
            if s_.__len__() > 1:
                s_prob = s_[0] / 1e6
                e_prob = e_[-1] / 1e6

            seizures_ += [
                {
                    'start_annotation': szr.start_uutc,
                    'end_annotation': szr.stop_uutc,
                    'prob_right': r,
                    'prob_left': l,
                    'self_reported': l.__len__() <= 1,
                    's_prob': s_prob,
                    'e_prob': e_prob,
                    'fs' : fs
                }
            ]
            
        self._close_sql()
        return seizures_


    def get_seizures_patient(self, patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"start_uutc, stop_uutc " \
                f"FROM {self._sql_db_name}.Annotations_Events WHERE id='{str(patient_id)}' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp} " \
                f"and (ev_type='aura' or ev_type='seizure') " \
                f"and source='patient'"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()
        df['start_uutc'] /= 1e6
        df['start_uutc'] += 10*60
        df['stop_uutc'] /= 1e6
        df['stop_uutc'] += 10*60
        return df


    def get_datarate(self,  patient_id=None, start_timestamp=None, stop_timestamp=None):
        if isinstance(start_timestamp, type(None)):
            start_timestamp = 0
        if isinstance(stop_timestamp, type(None)):
            stop_timestamp = datetime.now().timestamp()

        start_timestamp *= 1e6
        stop_timestamp *= 1e6

        query = f"SELECT " \
                f"start_uutc, value  " \
                f"FROM {self._sql_db_name}.Automated_Events WHERE id='{str(patient_id)}' " \
                f"and ev_type='data_rate' " \
                f"and start_uutc>={start_timestamp} " \
                f"and start_uutc<={stop_timestamp}"
        self._open_sql()
        df = pd.read_sql(query, self._sql_connection)
        self._close_sql()

        df['start_uutc'] /= 1e6
        df['start_uutc'] += 10*60
        return df




























# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import setuptools
from glob import glob

from setuptools import Command, Extension
import shlex
import subprocess
import os
import re

## get version from file
VERSIONFILE="./best/__init__.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))



setuptools.setup(
    name="best_toolbox",
    version=verstr,
    license='',
    url="https://github.com/mselair/best_toolbox",

    author="Filip Mivalt",
    author_email="mivalt.filip@mayo.edu",


    description="Python package designed for iEEG analysis and sleep classification.",
    long_description="Python package for EEG sleep classification and analysis. Developed by the laboratory of Bioelectronics Neurophysiology and Engineering - Mayo Clinic",
    long_description_content_type="",

    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        'best_toolbox': ["best/dbs/artifact_removal/_configs/*.yaml", "best/dbs/artifact_removal/_models/*.pt", "dbs/artifact_bank/*.mat"]
    },
    data_files=[
        ('dbs_artifact_removal_cfg', glob('best/dbs/artifact_removal/_configs/*.yaml')),
        ('dbs_artifact_removal_mdl', glob('best/dbs/artifact_removal/_models/*.pt')),
        ('dbs_artifact_bank_artifact', glob('best/dbs/artifact_bank/*.mat')),
    ],


    classifiers=[
        "Development Status :: 3 - Alpha",
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: POSIX :: Linux",
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ],
    python_requires='>=3.6',
    install_requires =[
        'numpy',
        'pandas',
        'scipy',
        'tqdm',
        'matplotlib',
        'seaborn',
        'scikit-learn',
        'umap-learn',
        'pymef',
        'mef_tools',
        'pyedflib',
        'h5py',
        'sqlalchemy',
        'PyMySQL',
        'pytz',
        'python-dateutil',
        'pyzmq',
        'sshtunnel',
        'torch',
        'pyyaml',
        'mne'
    ]
)









# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import pandas as pd
from copy import deepcopy

"""
Tools for correcting telemetry-based hypnograms
"""

def do_median_filtration(df):
    for k in range(1, df.__len__() - 1):
        if df.iloc[k - 1].annotation == df.iloc[k + 1].annotation and df.iloc[k].duration == 30:
            df.iloc[k]['annotation'] = df.iloc[k - 1].annotation
    return df


def fill_same_voids(df, time_threshold=5*60, initial_state='AWAKE'):
    new_df = []
    current_state = initial_state
    current_state_duration = time_threshold
    last_annotation_end = df.iloc[0]['start']

    for idx in range(df.__len__()):
        crow = df.iloc[idx]
        s_ = crow['start']
        e_ = crow['end']
        void = crow['start'] - last_annotation_end
        if void > 0 and void <= time_threshold and current_state == crow['annotation']:
            crow['start'] = last_annotation_end
            crow['duration'] = crow['end'] - crow['start']

        if crow['annotation'] == current_state:
            current_state_duration += crow['duration']
        else:
            current_state_duration = crow['duration']
            current_state = crow['annotation']

        last_annotation_end = e_
        new_df += [crow]
    return pd.DataFrame(new_df).reset_index(drop=True)


def fill_wakerem_voids(df, time_threshold=5*60, initial_state='AWAKE'):
    new_df = []
    current_state = initial_state
    current_state_duration = time_threshold
    last_annotation_end = df.iloc[0]['start']

    for idx in range(df.__len__()):
        crow = df.iloc[idx]
        s_ = crow['start']
        e_ = crow['end']
        void = crow['start'] - last_annotation_end
        if void > 0 and void <= time_threshold and current_state == 'AWAKE' and crow['annotation'] == 'REM':
            vrow = deepcopy(crow)
            vrow['annotation'] = 'AWAKE'
            vrow['start'] = last_annotation_end
            vrow['end'] = crow['start']
            vrow['duration'] = crow['end'] - crow['start']
            new_df += [vrow]

        if crow['annotation'] == current_state:
            current_state_duration += crow['duration']
        else:
            current_state_duration = crow['duration']
            current_state = crow['annotation']

        last_annotation_end = e_
        new_df += [crow]
    return pd.DataFrame(new_df).reset_index(drop=True)


def fill_nonrem_voids(df, time_threshold=5*60, initial_state='AWAKE'):
    new_df = []
    current_state = initial_state
    current_state_duration = time_threshold
    last_annotation_end = df.iloc[0]['start']

    for idx in range(df.__len__()):
        crow = df.iloc[idx]
        s_ = crow['start']
        e_ = crow['end']
        void = crow['start'] - last_annotation_end
        if void > 0 and void <= time_threshold and current_state in ('N2', 'N3', 'N') and crow['annotation'] in (
        'N2', 'N3', 'N'):
            vrow = deepcopy(crow)
            vrow['annotation'] = 'N'
            vrow['start'] = last_annotation_end
            vrow['end'] = crow['start']
            vrow['duration'] = crow['end'] - crow['start']
            new_df += [vrow]

        if crow['annotation'] == current_state:
            current_state_duration += crow['duration']
        else:
            current_state_duration = crow['duration']
            current_state = crow['annotation']

        last_annotation_end = e_
        new_df += [crow]
    return pd.DataFrame(new_df).reset_index(drop=True)


def fill_sleep_voids(df, time_threshold=5*60, initial_state='AWAKE'):
    new_df = []
    current_state = initial_state
    current_state_duration = time_threshold
    last_annotation_end = df.iloc[0]['start']

    for idx in range(df.__len__()):
        crow = df.iloc[idx]
        s_ = crow['start']
        e_ = crow['end']
        void = crow['start'] - last_annotation_end
        if void > 0 and void <= time_threshold and current_state != 'AWAKE' and crow['annotation'] != 'AWAKE':
            vrow = deepcopy(crow)
            vrow['annotation'] = 'SLP'
            vrow['start'] = last_annotation_end
            vrow['end'] = crow['start']
            vrow['duration'] = crow['end'] - crow['start']
            new_df += [vrow]

        if crow['annotation'] == current_state:
            current_state_duration += crow['duration']
        else:
            current_state_duration = crow['duration']
            current_state = crow['annotation']

        last_annotation_end = e_
        new_df += [crow]
    return pd.DataFrame(new_df).reset_index(drop=True)


def correct_rem(df, time_threshold=5*60, initial_state='AWAKE'):
    new_df = []
    current_state = initial_state
    current_state_duration = time_threshold
    last_annotation_end = df.iloc[0]['start']

    for idx in range(df.__len__()):
        crow = df.iloc[idx]
        s_ = crow['start']
        e_ = crow['end']
        void = crow['start'] - last_annotation_end
        if void <= 60 and current_state == 'AWAKE' and crow[
            'annotation'] == 'REM' and current_state_duration >= time_threshold:
            crow['annotation'] = 'AWAKE'

        if crow['annotation'] == 'AWAKE':
            current_state_duration += crow['duration']
        else:
            current_state_duration = 0
        current_state = crow['annotation']

        last_annotation_end = e_
        new_df += [crow]
    return pd.DataFrame(new_df).reset_index(drop=True)


def correct_hypnogram(df, time_threshold=60):
    new_df = deepcopy(df)
    new_df = fill_same_voids(new_df, time_threshold=time_threshold)
    new_df = fill_wakerem_voids(new_df, time_threshold=time_threshold)
    new_df = fill_nonrem_voids(new_df, time_threshold=time_threshold)
    new_df = fill_sleep_voids(new_df, time_threshold=time_threshold)
    new_df = correct_rem(new_df, time_threshold=time_threshold)
    # new_df = do_median_filtration(new_df)
    return new_df
















# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.




"""
A package for a standardized manipulation with hypnogram annotations, saving, loading across different file formats (CyberP{SG, NSRR, ...) but also
basic analysis of hypnograms. The package is based on pandas.DataFrame.
"""

















# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os
import pandas as pd
import xml.etree.ElementTree as ET
from best.types import TwoWayDict

class NSRRSleepFile:
    def __init__(self, path=None, nsrr2hypnogram_keys={
        'Wake|0': 'WAKE',
        'Stage 1 sleep|1': 'N1',
        'Stage 2 sleep|2': 'N2',
        'Stage 3 sleep|3': 'N3',
        'Stage 4 sleep|4': 'N3',
        'REM sleep|5': 'REM'
    }):
        self.namespaces = TwoWayDict() # list of namespaces - key=prefix; value=uri
        self.Element = None
        self.path = path
        self.strp_format = '%Y.%m.%d. %H:%M:%S'

        self.nsrr2hypnogram = TwoWayDict(nsrr2hypnogram_keys)



        self.epoch_length = None
        self.software_version = None
        self.events = None

        if path:
            if os.path.isfile(path):
                self.read_file(path_xml=path)

    def read_file(self, path_xml):
        self.Element = ET.parse(path_xml).getroot()
        for Element in self.Element:
            tag = Element.tag
            if tag == 'SoftwareVersion':
                self.software_version = Element.text.split('\n')[0]
            if tag == 'EpochLength':
                self.epoch_length = float(Element.text.split('\n')[0])

        self._parse_file()

    def _parse_file(self):
        events = []
        for Element in self.Element:
            tag = Element.tag
            if tag == 'ScoredEvents':
                for ScoredEvent in Element:
                    tag_ScoredEvent = ScoredEvent.tag
                    event = {}
                    for Property in ScoredEvent:
                        tag_Property = Property.tag
                        if Property.text:
                            text_Property = Property.text.split('\n')[0]
                            event[tag_Property] = text_Property
                    events += [event]
        self.events = events

    def get_hypnogram(self):
        hyp = pd.DataFrame([ev for ev in self.events if 'EventType' in ev.keys() if ev['EventType'] == 'Stages|Stages'])
        hyp['duration'] = [float(x) for x in hyp['Duration']]
        hyp = hyp.drop(['Duration'], axis=1)

        hyp['start'] = [float(x) for x in hyp['Start']]
        hyp = hyp.drop(['Start'], axis=1)

        hyp = hyp.drop(['EventType'], axis=1)

        hyp['annotation'] = [self.nsrr2hypnogram[an] for an in hyp['EventConcept']]
        hyp = hyp.drop(['EventConcept'], axis=1)

        hyp['end'] = hyp['start'] + hyp['duration']
        hyp = hyp[['annotation', 'start', 'end', 'duration']]
        return hyp

    def get_annotation_types(self):
        types = {}
        for Element in self.Element:
            tag = Element.tag
            if '}' in tag: tag = tag.split('}')[-1]
            if tag == 'AnnotationTypes':
                for AnnotationTypeElement in Element:
                    name = None
                    id = None
                    for SubElement in AnnotationTypeElement:
                        subtag = SubElement.tag
                        if '}' in subtag: subtag = subtag.split('}')[-1]
                        if subtag == 'id': id = SubElement.text
                        if subtag == 'name': name = SubElement.text
                    types[id] = name
        return types
